<!--
    VariableAnnotator

    Adds a Note element to each variable reference, of the following format:

    <Note kind="varRef" free="yes/no" assignable="yes/no" mutable="yes/no" scope-id="<id>">
        <Type ... />
    </Note>

    assumes: TagPorts, AddID

    author: JWJ
-->

<xsl:stylesheet
  xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
  version="1.1"
  xmlns:exsl="http://exslt.org/common"
  xmlns:set="http://exslt.org/sets"
  extension-element-prefixes="exsl set">

  <xsl:output method="xml"/>

    <xsl:template match="Actor">
        <xsl:copy>
            <xsl:for-each select="@*">
                <xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute>
            </xsl:for-each>

            <xsl:apply-templates select="*"/>
        </xsl:copy>
    </xsl:template>


    <xsl:template match="Expr[@kind='Var']">
        <xsl:variable name="name"><xsl:value-of select="@name"/></xsl:variable>

        <!--
            Compute list of all lexically enclosing decls of the name $name.
            Later, if that is not empty, choose the last.
            Decls are enumerated along the ancestor axis: innermost decl is last.
        -->
        <xsl:variable name="decls">
            <xsl:for-each select="./ancestor::*">
                <xsl:choose>
                    <xsl:when test="Decl[@kind='Variable'][@name=$name]">
                        <Note kind="varRef" free="no" assignable="{Decl/@assignable}"
                              mutable="{Decl/@mutable}" scope-id="{@id}">
                              <xsl:copy-of select="Decl[@name=$name]/Type"/>
                        </Note>
                    </xsl:when>
                    <xsl:when test="Decl[@kind='Parameter'][@name=$name]">
                        <Note kind="varRef" free="no" assignable="no" mutable="no" scope-id="{@id}">
                              <xsl:copy-of select="Decl[@name=$name]/Type"/>
                        </Note>
                    </xsl:when>
                    <xsl:when test="Input/Decl[@name=$name]">
                        <xsl:variable name="port"><xsl:value-of select="Input/@port"/></xsl:variable>
                        <Note kind="varRef" free="no" assignable="no" mutable="no" scope-id="{@id}">
                              <xsl:copy-of select="ancestor::*/Port[@name=$port][1]/Type"/>
                        </Note>
                    </xsl:when>
                    <xsl:when test="Input/Decl[@name=$name]">
                        <xsl:variable name="port"><xsl:value-of select="Input/@port"/></xsl:variable>
                        <Note kind="varRef" free="no" assignable="no" mutable="no" scope-id="{@id}">
                            <Type name="Seq">
                                <TypePars>
                                    <xsl:copy-of select="ancestor::*/Port[@name=$port][1]/Type"/>
                                </TypePars>
                            </Type>
                        </Note>
                    </xsl:when>
                </xsl:choose>
            </xsl:for-each>
        </xsl:variable>

        <xsl:copy>
            <xsl:for-each select="@*">
                <xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute>
            </xsl:for-each>
            <xsl:apply-templates select="*"/>

            <!--
                Pick the first (innermost) decl if there is one. Otherwise, it is a free variable.
            -->
            <xsl:choose>
                <xsl:when test="$decls/Note">
                    <!-- FIXME: Is this a bug in Saxon? The Note elements should have been created in reverse
                        document order, because we used the ancestor:: axis. So if this is a bug, the 'correct'
                        select in the next line would be:
                            exsl:node-set($decls)/Note[1]
                    -->
                    <xsl:copy-of select="$decls/Note[last()]"/>
                </xsl:when>
                <xsl:otherwise>
                    <Note kind="varRef" free="yes" assignable="no" mutable="no"/>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:copy>
    </xsl:template>


    <xsl:template match="*">
        <xsl:copy>
            <xsl:for-each select="@*">
                <xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute>
            </xsl:for-each>

            <xsl:apply-templates select="node() | text()"/>
        </xsl:copy>
    </xsl:template>

</xsl:stylesheet>





