// GarpActors.cpp
//
// each routine is an actor in the Garp pipeline
//
// Author:  Rich Williams
//

#include "Algorithm.h"
#include "EnvLayerSet.h"
#include "EnvLayer.h"
#include "RuleSet.h"

#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <cstdlib>
#include <iostream>

#include "org_ecoinformatics_seek_garp_GarpJniGlue.h"
#include <jni.h>

extern void PresampleLayers(char* s1, char* s2, char* s3);
extern void RunGarpAlgorithm(char* s1, char* s2);
extern void DoGarpPrediction(char* s1, char* s2, char* s3, char* s4);

//JNI function calls
JNIEXPORT void JNICALL Java_org_ecoinformatics_seek_garp_GarpJniGlue_PresampleLayers
(JNIEnv *env, jobject obj, jstring s1, jstring s2, jstring s3)
{
  std::cout << "Calling PresampleLayers in CPP Code.\n";
  jboolean iscopy;
  char * lfile = (char*) env->GetStringUTFChars(s1, &iscopy);
  char * dfile = (char*) env->GetStringUTFChars(s2, &iscopy);
  char * cfile = (char*) env->GetStringUTFChars(s3, &iscopy);
  std::cout << "calling presample layers now\n";
  PresampleLayers(lfile, dfile, cfile);
  std::cout << "DONE Calling PresampleLayers in CPP Code.\n";
}
JNIEXPORT void JNICALL Java_org_ecoinformatics_seek_garp_GarpJniGlue_RunGarpAlgorithm
    (JNIEnv *env, jobject obj, jstring s1, jstring s2)
{
  std::cout << "Calling RunGarpAlgorithm in CPP Code.\n";
  jboolean iscopy;
  char * cfile = (char*) env->GetStringUTFChars(s2, &iscopy);
  char * rfile = (char*) env->GetStringUTFChars(s1, &iscopy);
  RunGarpAlgorithm(cfile, rfile);
  std::cout << "DONE Calling RunGarpAlgorithm in CPP Code.\n";
}

JNIEXPORT void JNICALL Java_org_ecoinformatics_seek_garp_GarpJniGlue_DoGarpPrediction
(JNIEnv *env, jobject obj, jstring s1, jstring s2, jstring s3, jstring s4)
{
  std::cout << "Calling DoGarpPrediction in CPP Code.\n";
  jboolean iscopy;
  char * rfile = (char*) env->GetStringUTFChars(s1, &iscopy);
  char * lfile = (char*) env->GetStringUTFChars(s2, &iscopy);
  char * afile = (char*) env->GetStringUTFChars(s3, &iscopy);
  char * bfile = (char*) env->GetStringUTFChars(s4, &iscopy);

    DoGarpPrediction(rfile, lfile, afile, bfile);

  std::cout << "DONE Calling DoGarpPrediction in CPP Code.\n";
}



// input:  layerSetFileName (XML file, the environmental layers)
// input:  dataPointFileName (the species occurrence data)
// output:  cellSetFileName (XML file, the sampled layers and species occurrences)
//
void PresampleLayers(char* layerSetFileName, char* dataPointFileName,
  char* cellSetFileName)
{
  std::cout << "1";
  std::cout << "layerSetFileName: " << layerSetFileName << "\n";
  std::cout << "dataPointFileName: " << dataPointFileName << "\n";
  std::cout << "cellSetFileName: " << cellSetFileName << "\n";
  std::cout << "2";
  char * sDataPoints, sLine[2048], sLong[256], sLat[256];
  FILE * fpoints;
  int iSize;

  GarpUtil::randomize(0);

  iSize = getFileSize((char*)dataPointFileName);

  fpoints = fopen((char*)dataPointFileName, "r");
  if (fpoints)
  {
    // read data points from file
    sDataPoints = new char[iSize + 1];
    strcpy(sDataPoints, "");
    while (!feof(fpoints))
    {
      fscanf(fpoints, "%s\t%s\n", sLong, sLat);
      sprintf(sLine, "%s %s ", sLong, sLat);
      strcat(sDataPoints, sLine);
    }

    fclose(fpoints);
  }


  EnvLayerSet oEnvLayerSet;
  EnvCellSet oTrainSet;
  // load env layer set from dxl file
  oEnvLayerSet.fromXMLFile((char*)layerSetFileName, 2);
  // initialize training cell set to hold at most 2500 resampled points
  oTrainSet.initialize(2500);

  // resample points from datapoints read from file
  oEnvLayerSet.presampleFromDataPoints(&oTrainSet, sDataPoints, true, 2500, -1);
  oTrainSet.toXMLFile((char*)cellSetFileName, (char*)0);
  std::cout << "****done presampling layers****\n";
}

// input: cellSetFileName (XML, occurence and environmental data samples)
// output: ruleSetFileName (XML, the rule set)
//
void RunGarpAlgorithm(char* cellSetFileName, char* ruleSetFileName)
{
  std::cout << "cellSetFileName: " << cellSetFileName << "\n";
  std::cout << "ruleSetFileName: " << ruleSetFileName << "\n";

  GarpAlgorithm oGarp;
  EnvCellSet oTrainSet;
  RuleSet * pRuleSet;

  GarpUtil::randomize(0);

  //takes xml string not filename (there's no toXMLFile method in this object)
  oTrainSet.fromXMLFile((char*)cellSetFileName);
  // initialize oGarp algorithm
  oGarp.setParameter("Totalgens", "20");
  //TODO: NEED TO MAKE GarpAlgorithm work with no layer set
  //takes layerset and training set
  oGarp.initialize(0, &oTrainSet);
  //oGarp.setSelectedLayers("5 6 7");

  // main algorithm loop
  while (!oGarp.done())
  {
    oGarp.iterate();
    std::cout << "[Generation: " << oGarp.getParameter("Gen") << "]" <<
      "[Convergence: " << oGarp.getParameter("Convergence") << "]\n";
  }

  std::cout << "done with main garp loop\n";

  // get a pointer to internal rule set. will be deallocated by oGarp object
  pRuleSet = oGarp.getBestSet();
  pRuleSet->toXMLFile((char*)ruleSetFileName, "Best");

  std::cout << "****done with main garp algorithm*****\n";
}

// input: ruleSetFileName (XML, a GARP rule set)
// input: layerSetFileName (XML, the layers to predict over)
// output: ascii file
// output: bitmap file
//
void DoGarpPrediction(char* ruleSetFileName, char* layerSetFileName,
  char* outputASCIIFile, char* outputBMPFile)
{
  std::cout << "****in DoGarpPrediction cpp code****\n";
  RuleSet oRuleSet;
  EnvLayer oResultLayer, oSpeciesLayer;
  EnvLayerSet oEnvLayerSet;

  GarpUtil::randomize(0);
  oRuleSet.fromXMLFile((char*)ruleSetFileName);
  oEnvLayerSet.fromXMLFile((char*)layerSetFileName, 2);
  oRuleSet.setModelLayerSet(&oEnvLayerSet);
    // save result to bitmap and ASCII raster grid
  oEnvLayerSet.metadataToEnvLayer(&oResultLayer);
  oResultLayer.createMatrix("RawByteMatrixInMemory", "");
  oResultLayer.asSpecies();
  oRuleSet.predict(&oResultLayer, &oEnvLayerSet, 0.0);
  oResultLayer.toAsciiGrid((char*)outputASCIIFile);

  oResultLayer.toBitmap((char*)outputBMPFile);
  std::cout << "****done with DoGarpPrediction*****\n";
}

