/**
 *    '$RCSfile: DataCacheItem.java,v $'
 *
 *     '$Author: rspears $'
 *       '$Date: 2004/10/13 20:16:43 $'
 *   '$Revision: 1.8 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
package org.ecoinformatics.seek.datasource;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;

import org.ecoinformatics.seek.querybuilder.DBUIUtils;
import org.w3c.dom.Node;

/**
 * Represents an item in the cache
 */
public abstract class DataCacheItem implements Runnable
{
    public static final int CACHE_OK       = 0;
    public static final int CACHE_BUSY     = 1;
    public static final int CACHE_ERROR    = 2;
    public static final int CACHE_COMPLETE = 3;
    public static final int CACHE_EMPTY    = 4;
    
    protected org.ecoinformatics.util.DebugPrinter dbg;
    
    public    Thread mThread        = null;
    protected byte[] mData          = null;
    protected Date   mCreatedDate   = new Date();
    protected String mName          = null;
    protected String mResourceName  = null;
    protected String mLocalFileName = null;
    protected int    mSize          = 0;
    
    protected Vector  mListeners    = new Vector();
    protected Vector  mNotifierList = null;
    
    protected boolean mIsLoading    = false;
    protected int     mCurrInx      = 0;
    protected int     mStatus       = CACHE_EMPTY;
    protected boolean mHasChanged   = true;
    
    /**
     * Constructor 
     *
     */
    public DataCacheItem()
    {
        dbg = org.ecoinformatics.util.DebugPrinter.getInstance();
    }
    
   
    /**
     * Add a listenero 
     * @param aListener the listener to add
     */
    public void addListener(DataCacheListener aListener)
    {
        if (aListener != null)
        {
            mListeners.addElement(aListener);
        }
    }
    
    /**
     * Removes a listener 
     * @param aListener the listener to remove
     */
    public void removeListener(DataCacheListener aListener)
    {
        if (aListener != null)
        {
            mListeners.removeElement(aListener);
        }
    }
    
    /**
     * Removes all the listeners
     */
    public void removeAllListeners()
    {
        mListeners.clear();
    }
    
    /**
     * Notifies the listener that the "getting" of the data has completed
     *
     */
    public synchronized void notifyListeners()
    {
        mNotifierList = new Vector(mListeners);
        
        for (Enumeration e = mNotifierList.elements(); e.hasMoreElements();)
        {
            DataCacheListener l = (DataCacheListener)e.nextElement();
            //System.err.println("notifyListeners: "+this + "  " +l);
            l.complete(this);
        }
    }
    
    /**
     * Returns the data as a byte array
     * @return
     */
    public byte[] getData()
    {
        if (mData == null)
        {
          loadData();
        }
        return mData;
    }
    
    /**
     * Sets the data into the cache item
     * @param aData the new data byte array
     */
    public void setData(byte[] aData)
    {
        mData = aData;
        if (mData != null)
        {
            mSize = mData.length;
        } 
        mHasChanged = true;
    }
    
    /**
     * Return the status of getting the data
     * @return
     */
    public boolean isReady()
    {
        return mStatus == CACHE_OK || mStatus == CACHE_COMPLETE;
    }
    
    /**
     * Returns whether it is busy getting the data
     * @return
     */
    public boolean isBusy()
    {
        return mStatus == CACHE_BUSY;
    }
    
    /**
     * Refreshes the data from the original source
     * @param aListener
     */
    public void refresh(DataCacheListener aListener)
    {
        clear();
        mLocalFileName = createFileName();
        addListener(aListener);
        mCreatedDate = new Date();
        mData = null;
        mHasChanged = true;
        start();
    }
    
    public void reset()
    {
        clear();
        mStatus = CACHE_EMPTY;
        mLocalFileName = null;
    }
    
    /**
     * Clear the dat from this cache item
     *
     */
    public void clear()
    {
        if (mLocalFileName != null && mLocalFileName.length() > 0)
        {
            File file = new File(mLocalFileName);
            if (file != null && file.exists())
            {
                file.delete();
            }
        }
        mData = null;
        removeAllListeners();
    }
    
    /**
     * Abstract method for actually getting the data while in the thread
     *
     */
    public abstract void   doWork();
    
    /**
     * Returns the class name of the object that should be created to handle this type of data
     * @return
     */
    protected abstract String getSerialClassName();
    
    /**
     * Returns any extra attributes a derived class may have that needs to augment the XML
     * @return
     */
    protected abstract String getAttrsForXML();
    
   
    /**
     * Returns the XML representing this item
     * @return
     */
    public String toXMLString()
    {
        StringBuffer strBuf = new StringBuffer();
        if (mLocalFileName != null && 
            mData != null && 
            mSize > 0 && 
            mName.length() > 0 &&
            mResourceName.length() > 0)
        {
           strBuf.append("  <item\n");
           strBuf.append("   classname=\"" + getSerialClassName() +"\"\n");
           strBuf.append("   name=\"" + mName +"\"\n");
           strBuf.append("   resname=\"" + mResourceName +"\"\n");
           strBuf.append("   date=\"" + (mCreatedDate != null ? mCreatedDate.getTime() : 0) +"\"\n");
           strBuf.append("   fileName=\"" + mLocalFileName +"\"\n");
           strBuf.append("   " + getAttrsForXML());
           strBuf.append("   size=\"" + mData.length +"\"");
           strBuf.append("/>\n");
        }
       return strBuf.toString();
    }
    
    /**
     * FIlls in this node from an XML Node in the DOM tree
     * @param aNode the DOM node representing this object
     */
    public boolean buildFromDOM(Node aNode)
    {
        
        if (aNode != null) 
        {
            mName          = DBUIUtils.findAttrValue(aNode, "name");
            mResourceName  = DBUIUtils.findAttrValue(aNode, "resname");
            mLocalFileName = DBUIUtils.findAttrValue(aNode, "fileName");
            try {
                long time = Long.parseLong(DBUIUtils.findAttrValue(aNode, "date"));
                mCreatedDate = new Date(time);
            } catch (Exception e) 
            {
                return false;
            }
            try {
                mSize = Integer.parseInt(DBUIUtils.findAttrValue(aNode, "size"));
            } catch (Exception e) 
            {
                return false;
            }
        }
        mHasChanged = false;
        return true;
    }
    
    /**
     * Loads the data from the cache (a file)
     * @return whether it was able to load the data
     */
    public boolean loadData()
    {
        if (mLocalFileName != null && mLocalFileName.length() > 0)
        {
            try 
            {
                File file = new File(mLocalFileName);
                if (file != null && file.exists())
                {
                    long size = file.length();
                    if (size <= Integer.MAX_VALUE) // XXX this could be bad for large datasets
                    {
                        int iSize = (int)size; 
                        if (mSize > 0)
                        {
                            if (mData == null || iSize != mSize)
                            {
                              mData = new byte[mSize];
                            }
                        } else {
                            dbg.print("loadData() - mSize == 0", 2);
                            return false;
                        }
                        FileInputStream fis = new FileInputStream(file);
                        if (fis != null)
                        {
                            BufferedInputStream bis = new BufferedInputStream(fis);
                            if (bis != null)
                            {
                                int retSize = bis.read(mData, 0, iSize);
                                if (retSize != iSize) 
                                {
                                    dbg.print("loadData - Wrong amount of data read: \n  " + mName+"  \n"+mResourceName+"  \n"+mLocalFileName, 2);            
                                }
                                bis.close();
                                mStatus = CACHE_OK;
                                dbg.print("DataCacheItem - Data was loaded: " + mName+"  \n"+mResourceName+"  \n"+mLocalFileName, 2);
                            }
                        }
                    } else 
                    {
                        dbg.print("loadData - Too Much Data for byte array: \n  " + mName+"  \n"+mResourceName+"  \n"+mLocalFileName, 2);            
                    }
                }
                
            } catch (Exception e)
            {
                System.err.println(e);
                return false;
            }
            return true;
        } else 
        {
            dbg.print("loadData - mLocalFileName was null: \n  " + mName+"  \n"+mResourceName+"  \n"+mLocalFileName, 2);            
            return false;
        }
    }
    
    /**
     * Creates a new unique data file based on the time
     * @return the name of the file
     */
    protected String createFileName()
    {
        
        String fileName = DataCacheManager.getInstance().getDataPath();
        fileName += Long.toString(mCreatedDate.getTime()) + ".dat";
        File file = new File(fileName);
        return file.getAbsolutePath();
    }
    
    /**
     * Save the data out to a file
     * @return
     */
    public boolean saveData()
    {
        if (mData != null && mHasChanged)
        {
            try 
            {
                if (mLocalFileName == null || mLocalFileName.length() == 0) 
                {
                    mLocalFileName = createFileName();
                }
                
                if (mData != null && mData.length > 0)
                {
                    File file = new File(mLocalFileName);
                    FileOutputStream fos = new FileOutputStream(file);
                    BufferedOutputStream bos = new BufferedOutputStream(fos);
                    bos.write(mData);
                    bos.close();
                    mHasChanged = false;
                }
                
            } catch (Exception e)
            {
                System.err.println(e);
                return false;
            }
        }
        return true;
    }
    
    //----------------------------------------------------------------
    //-- Runnable Interface
    //----------------------------------------------------------------
    public void start() 
    {
        if (mThread == null) 
        {
            mThread = new Thread(this);
            mThread.setPriority(Thread.MIN_PRIORITY);
            mThread.setName(mName);
            mThread.start();
        }
    }

    /**
     * 
     *
     */
    public synchronized void stop() 
    {
        if (mThread != null) 
        {
            mThread.interrupt();
        }
        mThread = null;
        notifyAll();
    }

    /**
     * 
     */
    public void run() 
    {
        Thread me = Thread.currentThread();

        if (!isReady())
        {
            mStatus = CACHE_BUSY;
            try 
            {
                Thread.sleep(200);
            } catch (InterruptedException e) { }
            
            while (mThread == me && mStatus == CACHE_BUSY) 
            {
                // do work here 
                doWork();
                if (mStatus == CACHE_BUSY)
                {
                    try 
                    {
                        Thread.sleep(200);
                    } catch (InterruptedException e) { }
                }
            }
        }
        
        saveData();
   
        dbg.print("run - Done With Work.", 2);            
        notifyListeners();
        
        mThread = null;
    }
    /**
     * @return Returns the mCreatedDate.
     */
    public Date getCreatedDate()
    {
        return mCreatedDate;
    }
    /**
     * @param createdDate The mCreatedDate to set.
     */
    public void setCreatedDate(Date createdDate)
    {
        mCreatedDate = createdDate;
    }
    /**
     * @return Returns the mLocalFileName.
     */
    public String getLocalFileName()
    {
        return mLocalFileName;
    }
    
    /**
     * @return Returns the mLocalFileName.
     */
    public String getFileName()
    {
        int inx = mLocalFileName.lastIndexOf('/');
        if (inx > -1) 
        {
          return mLocalFileName.substring(inx+1);
        }
        return mLocalFileName;
    }
    
    /**
     * @param localFileName The mLocalFileName to set.
     */
    public void setLocalFileName(String localFileName)
    {
        mLocalFileName = localFileName;
    }
    /**
     * @return Returns the mName.
     */
    public String getName()
    {
        return mName;
    }
    /**
     * @param name The mName to set.
     */
    public void setName(String name)
    {
        mName = name;
    }
    
    /**
     * @return Returns the mResourceName.
     */
    public String getResourceName()
    {
        return mResourceName;
    }
    
    /**
     * @param resourceName The mResourceName to set.
     */
    public void setResourceName(String resourceName)
    {
        mResourceName = resourceName;
    }
    
    /**
     * @return Returns the mStatus.
     */
    public int getStatus()
    {
        return mStatus;
    }
    /**
     * @return Returns the mSize.
     */
    public int getSize()
    {
        return mSize;
    }
}
