/**
 *    '$RCSfile: Eml200Parser.java,v $'
 *
 *     '$Author: tao $'
 *       '$Date: 2004/09/04 00:03:06 $'
 *   '$Revision: 1.10 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.datasource.eml.eml2;

import java.io.InputStream;
import java.util.Hashtable;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.xpath.XPathAPI;
import org.ecoinformatics.seek.datasource.Attribute;
import org.ecoinformatics.seek.datasource.DataTypeResolver;
import org.ecoinformatics.seek.datasource.Domain;
import org.ecoinformatics.seek.datasource.TableEntity;
import org.ecoinformatics.seek.datasource.UnresolvableTypeException;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * This plugin parses EML 2.0.0 metadata files
 */
public class Eml200Parser
{

    private static String NAMESPACE = "eml://ecoinformatics.org/eml-2.0.0";
    private Hashtable entityHash = new Hashtable();
    private Hashtable fileHash = new Hashtable();
    private int numEntities = -1;
    private int numRecords = -1;
    private TableEntity entityObject = null;
    private DataTypeResolver dtr = DataTypeResolver.instanceOf();
    private String elementId = "0";

    /**
     * returns a hashtable of with the id of the entity as the key and the data
     * file id to which the entity refers as the value. This way, if you want to
     * know what data file goes with an entity, you can do a get on this hash
     * for the id of the entity. note that the entity id is the xml entity id
     * from the generated input step, not the id of the entity file itself.
     */
    public Hashtable getDataFilesHash()
    {
        return fileHash;
    }

    /**
     * parses the EML package using an InputSource
     */
    public void parse(InputSource source) throws Exception
    {
        DocumentBuilder builder = DocumentBuilderFactory.newInstance()
                        .newDocumentBuilder();
        Document doc = builder.parse(source);
        parseDocument(doc);
    }

    /**
     * parses the EML package using an InputStream
     */
    public void parse(InputStream is) throws Exception
    {
        DocumentBuilder builder = DocumentBuilderFactory.newInstance()
                        .newDocumentBuilder();
        Document doc = builder.parse(is);
        parseDocument(doc);
    }

    /**
     * parses the EML document
     */
    public void parseDocument(Document doc) throws Exception
    {
        NodeList entities;
        try {
            XPathAPI xpath = new XPathAPI();
            entities = XPathAPI.selectNodeList(doc, "//dataset/dataTable");
        } catch (Exception e) {
            throw new Exception("Error extracting entities from "
                            + "eml2.0.0 package.");
        }

        try {
            System.out.println("Processing entities");
            processEntities(entities);
            System.out.println("Done processing entities");
        } catch (Exception e) {
            throw new Exception("Error processing entities: " + e.getMessage());
        }
    }

    /**
     * returns a hashtable of entity names hashed to the entity description
     * metadata that goes with each entity.
     */
    public Hashtable getEntityHash()
    {
        return entityHash;
    }

    /**
     * returns the number of records in this dataItem
     *
     * @param entityId the id of the entity object to get the record count for
     */
    public int getRecordCount(String entityId)
    {
        return ((TableEntity) entityHash.get(entityId)).getNumRecords();
    }

    /**
     * returns the total number of entities in the data item collection that was
     * passed to this class when the object was created.
     */
    public int getEntityCount()
    {
        return numEntities;
    }

    /**
     * returns the number of attributes in the given entity
     *
     * @param entityId the id of the entity object that you want the attribute
     *            count for
     */
    public int getAttributeCount(String entityId)
    {
        Attribute[] attArray = ((TableEntity) entityHash.get(entityId))
                        .getAttributes();
        return attArray.length;
    }

    /**
     * process the attributes
     */
    private void processAttributes(NodeList atts, TableEntity entObj)
                    throws Exception
    {
        for (int i = 0; i < atts.getLength(); i++) { //go through each
                                                     // attribute
            Node att = atts.item(i);
            NodeList attChildren = att.getChildNodes();
            NamedNodeMap attAttributes = att.getAttributes();

            String attName = "";
            String attLabel = "";
            String attDefinition = "";
            String attUnit = "";
            String attUnitType = "";
            String attStorageType = "";
            String attMeasurementScale = "";
            String attPrecision = "";
            Domain domain = null;

            elementId = new Integer((new Integer(elementId)).intValue() + 1)
                            .toString();

            for (int j = 0; j < attChildren.getLength(); j++) {
                Node child = attChildren.item(j);
                String childName = child.getNodeName();
                if (childName.equals("attributeName")) {
                    attName = child.getFirstChild().getNodeValue().trim();
                } else if (childName.equals("attributeLabel")) {
                    attLabel = child.getFirstChild().getNodeValue().trim();
                } else if (childName.equals("attributeDefinition")) {
                    attDefinition = child.getFirstChild().getNodeValue().trim();
                } else if (childName.equals("measurementScale")) {
                    //unit is tricky because it can be custom or standard
                    Vector info = new Vector();
                    int domainType = Domain.DOM_NONE;
                    NodeList msNodeList = child.getChildNodes();
                    for (int k = 0; k < msNodeList.getLength(); k++) {
                        Node n = msNodeList.item(k);
                        String name = n.getNodeName();
                        if (name.equals("interval") || name.equals("ratio")) {
                            Node sUnit = XPathAPI.selectSingleNode(n,
                                            "unit/standardUnit");
                            Node cUnit = XPathAPI.selectSingleNode(n,
                                            "unit/customUnit");
                            if (sUnit != null) {
                                attUnit = sUnit.getFirstChild().getNodeValue();
                                attUnitType = Attribute.STANDARDUNIT;
                            } else if (cUnit != null) {
                                attUnit = cUnit.getFirstChild().getNodeValue();
                                attUnitType = Attribute.CUSTOMUNIT;
                            } else {
                                System.out.println("xpath didn't work");
                            }
                            Node precision = XPathAPI.selectSingleNode(n,
                                            "precision");
                            attPrecision = precision.getFirstChild()
                                            .getNodeValue();
                            Node dNode = XPathAPI.selectSingleNode(n,
                                            "numericDomain").getFirstChild();
                            String dName = dNode.getNodeName();
                            if (dName.equals("numberType")) // got number type
                            {
                                domainType = Domain.DOM_NUMBERTYPE;
                                info.add(dNode.getNodeValue());
                            } else if (dName.equals("boundsGroup"))
                                // got bounds group
                            {
                                NodeList boundsList = XPathAPI.selectNodeList(
                                                dNode, "./bounds");
                                for (i = 0; i < boundsList.getLength(); i++) {
                                    NodeList nl;
                                    Node bound;
                                    String min = "", max = "";
                                    String exclMin = null, exclMax = null;
                                    try {
                                        nl = XPathAPI.selectNodeList(boundsList
                                                        .item(i), "./minimum");
                                        bound = nl.item(0);
                                        min = bound.getFirstChild()
                                                        .getNodeValue();
                                        exclMin = bound.getAttributes().getNamedItem(
                                                  "exclusive").getNodeValue();
                                    } catch (Exception e) {
                                    }
                                    try {
                                        nl = XPathAPI.selectNodeList(boundsList
                                                        .item(0), "./maximum");
                                        bound = nl.item(0);
                                        max = bound.getFirstChild().getNodeValue();
                                        exclMax = bound.getAttributes().getNamedItem(
                                                  "exclusive").getNodeValue();
                                    } catch (Exception e) {
                                    }
                                    info.add(min);
                                    info.add(exclMin);
                                    info.add(max);
                                    info.add(exclMax);
                                    domainType = Domain.DOM_BOUNDED;
                                }
                            }
                        } else if (name.equals("nominal")
                                        || name.equals("ordinal")) {
                            Node dNode = XPathAPI.selectSingleNode(n,
                                            "nonNumericDomain").getFirstChild();
                            String dName = dNode.getNodeName();
                            if (dName.equals("textDomain")) {
                                NodeList nl = XPathAPI.selectNodeList(dNode,
                                                "./pattern");
                                for (int l = 0; l < nl.getLength(); l++) {
                                    info.add(nl.item(l).getFirstChild()
                                                    .getNodeValue());
                                }
                                domainType = Domain.DOM_TEXT;
                            } else if (dName.equals("enumeratedDomain")) {
                                NodeList nl = XPathAPI.selectNodeList(dNode,
                                                "./codeDefinition");
                                for (int l = 0; l < nl.getLength(); l++) {
                                    info.add(nl.item(l).getFirstChild()
                                                    .getNodeValue());
                                }
                                domainType = Domain.DOM_ENUMERATED;
                            }
                        } else if (name.equals("datetime")) {
                            info.add(XPathAPI.selectSingleNode(n,
                                            "./formatString").getNodeValue());
                            domainType = Domain.DOM_DATETIME;
                        }
                    }
                    domain = new Domain(domainType, info, null);
                } else if (childName.equals("storageType")) {
                    attStorageType = child.getFirstChild().getNodeValue()
                                    .trim();
                }
            }

            String resolvedType;
            try {
                resolvedType = dtr
                                .resolveType(null, null, null, attStorageType);
            } catch (UnresolvableTypeException ute) {
                throw new Exception("The type " + attStorageType
                                + " could not be resolved.");
            }

            Attribute attObj = new Attribute(elementId, attName, attLabel,
                            attDefinition, attUnit, attUnitType, resolvedType,
                            attMeasurementScale, attPrecision, domain);
            entObj.add(attObj);
        }
    }

    /**
     * pulls the entity information out of the xml and stores it in a hashtable
     */
    private void processEntities(NodeList entities) throws SAXException,
                    javax.xml.transform.TransformerException, Exception
    {
        numEntities = entities.getLength();
        String entityName = "";
        String entityDescription = "";
        String entityOrientation = "";
        String entityCaseSensitive = "";
        String entityNumberOfRecords = "-1";
        String physicalFile = "";
        String numHeaderLines = "0";
        String fieldDelimiter = "";
        String recordDelimiter = "";

        for (int i = 0; i < numEntities; i++) {
            //go through the entities and put the information into the hash.
            elementId = new Integer((new Integer(elementId)).intValue() + 1)
                            .toString();
            Node entity = entities.item(i);
            NodeList entityChildren = entity.getChildNodes();
            for (int j = 0; j < entityChildren.getLength(); j++) {
                Node child = entityChildren.item(j);
                String childName = child.getNodeName();

                if (childName.equals("entityName")) {
                    entityName = child.getFirstChild().getNodeValue();
                } else if (childName.equals("entityDescription")) {
                    entityDescription = child.getFirstChild().getNodeValue();
                } else if (childName.equals("caseSensitive")) {
                    entityCaseSensitive = child.getFirstChild().getNodeValue();
                } else if (childName.equals("numberOfRecords")) {
                    entityNumberOfRecords = child.getFirstChild()
                                    .getNodeValue();
                    numRecords = (new Integer(entityNumberOfRecords))
                                    .intValue();
                }

                NodeList orientationNodeList = XPathAPI.selectNodeList(entity,
                        "physical/dataFormat/textFormat/attributeOrientation");
                if (orientationNodeList != null) {
                    entityOrientation = orientationNodeList.item(0).getFirstChild().getNodeValue();
                }

                NodeList headerLinesNL = XPathAPI.selectNodeList(entity,
                        "physical/dataFormat/textFormat/numHeaderLines");
                if ((headerLinesNL != null) && (headerLinesNL.getLength() > 0)) {
                    Node headerLinesNode = headerLinesNL.item(0);
                    if (headerLinesNode != null) {
                        numHeaderLines = headerLinesNode.getFirstChild()
                                        .getNodeValue();
                    }
                }

                NodeList delimiterNL = XPathAPI.selectNodeList(entity,
                        "physical/dataFormat/textFormat/simpleDelimited/fieldDelimiter");
                if (delimiterNL != null) {
                    fieldDelimiter = delimiterNL.item(0).getFirstChild().getNodeValue();
                }

                NodeList recDelimiterNL = XPathAPI.selectNodeList(entity,
                        "physical/dataFormat/textFormat/recordDelimiter");
                if ((recDelimiterNL != null) && (recDelimiterNL.getLength() > 0)) {
                    recordDelimiter = recDelimiterNL.item(0).getFirstChild().getNodeValue();
                } else {
                    recordDelimiter = "\r\n";
                }

                //get the distribution information
                NodeList distributionNL = XPathAPI.selectNodeList(entity,
                                "physical/distribution/online/url");
                physicalFile = distributionNL.item(0).getFirstChild()
                                .getNodeValue();
            }

            if (entityOrientation.trim().equals("column")) {
                entityOrientation = TableEntity.COLUMNMAJOR;
            } else {
                entityOrientation = TableEntity.ROWMAJOR;
            }

            if (entityCaseSensitive.equals("yes")) {
                entityCaseSensitive = "true";
            } else {
                entityCaseSensitive = "false";
            }

            entityObject = new TableEntity(elementId, entityName.trim(),
                            entityDescription.trim(), new Boolean(
                                            entityCaseSensitive),
                            entityOrientation, new Integer(
                                            entityNumberOfRecords).intValue());
            entityObject.setNumHeaderLines((new Integer(numHeaderLines))
                            .intValue());
            entityObject.setDelimiter(fieldDelimiter);
            entityObject.setRecordDelimiter(recordDelimiter);
            entityObject.setURL(physicalFile);
            try {
                NodeList attNL = XPathAPI.selectNodeList(entity,
                                "attributeList/attribute");
                processAttributes(attNL, entityObject);
            } catch (Exception e) {
                throw new Exception("Error parsing attributes: "
                                + e.getMessage());
            }

            entityHash.put(elementId, entityObject);
            fileHash.put(elementId, physicalFile);
        }
    }
}
