/**
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.sdm.spa;

import java.util.Vector;
import java.io.ByteArrayInputStream;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.traversal.NodeIterator;
import org.apache.xpath.XPathAPI;

import ptolemy.actor.TypedAtomicActor;
import ptolemy.actor.TypedIOPort;
import ptolemy.actor.lib.Const;
import ptolemy.data.ArrayToken;
import ptolemy.data.BooleanToken;
import ptolemy.data.RecordToken;
import ptolemy.data.StringToken;
import ptolemy.data.IntToken;
import ptolemy.data.expr.Parameter;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Workspace;

//////////////////////////////////////////////////////////////////////////
//// EnumHomologs
/**
   Extract homologs and send them as an ArrayToken, where
   each element is a RecordToken representing one homolog.

   @author Xiaowen Xin
   @version $Revision: 1.5 $
*/

public class EnumHomologs extends TypedAtomicActor {

    /** Construct the class with the given container and name.
     *  @param container The container.
     *  @param name The name of this actor.
     *  @exception IllegalActionException If the entity cannot be contained
     *   by the proposed container.
     *  @exception NameDuplicationException If the container already has an
     *   actor with this name.
     */
    public EnumHomologs(CompositeEntity container, String name)
        throws NameDuplicationException, IllegalActionException {
        super(container, name);

        portInput = new TypedIOPort(this, "input", true, false);
        portInput.setTypeEquals(BaseType.STRING);

        portOutput = new TypedIOPort(this, "output", false, true);

        _attachText(
            "_iconDescription",
            "<svg>\n"
                + "<rect x=\"0\" y=\"0\" "
                + "width=\"60\" height=\"20\" "
                + "style=\"fill:white\"/>\n"
                + "</svg>\n");
    }

    ///////////////////////////////////////////////////////////////////
    ////                     ports and parameters                  ////

    public TypedIOPort portInput;
    public TypedIOPort portOutput;

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Clone the actor into the specified workspace.
     *  @param workspace The workspace for the new object.
     *  @return A new actor.
     *  @exception CloneNotSupportedException If a derived class contains
     *   an attribute that cannot be cloned.
     */
    /*public Object clone(Workspace workspace)
        throws CloneNotSupportedException {
        EnumHomologs newObject = (EnumHomologs) super.clone(workspace);

        // Set the type constraint.
        newObject.portInput.setTypeEquals(BaseType.STRING);
        return newObject;
    }


    /** Parse the input string and output an ArrayToken where each element
     *  is a RecordToken representing the homolog.
     *  @exception IllegalActionException If it is thrown by the
     *   send() method sending out the token.
     *  @exception javax.xml.transform.TransformerException If it
     *   is thrown by parser methods.
     */
    public void fire() throws IllegalActionException {

        String strInput = ((StringToken) (portInput.get(0))).stringValue();

        _debug("input: ");
        _debug(strInput);

        //Following pathes directly query the ncbi blast result.
        String p_cpath  = "/BlastOutput/BlastOutput_iterations/Iteration/Iteration_hits/Hit";
        String p_xpath_gi = "Hit_id";
        String p_xpath_start = "Hit_hsps/Hsp/Hsp_hit-from";
        String p_xpath_end = "Hit_hsps/Hsp/Hsp_hit-to";
        String p_xpath_query_start = "Hit_hsps/Hsp/Hsp_query-from";

        // get rid of the doctype
        strInput = strInput.replaceFirst("\\<\\!DOCTYPE.+?\\>", "");

        Document doc;

        try {
            // parse input
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            doc = db.parse(new ByteArrayInputStream(strInput.getBytes()));

        } catch(Exception e) {
            System.err.println("EnumHomologs: could not parse input.");
            e.printStackTrace();
            throw new IllegalActionException("EnumHomologs: could not parse input.");
        }

        // get list of all nodes representing hits
        Node nodeRoot = doc.getDocumentElement().getFirstChild();
        NodeIterator nodesHits = null;
        Node node = null;
        
        try {
            nodesHits = XPathAPI.selectNodeIterator(nodeRoot, p_cpath);
        } catch(javax.xml.transform.TransformerException e) {
            System.err.println("EnumHomologs: could not select nodes.");
            e.printStackTrace();
            throw new IllegalActionException("EnumHomologs: could not select nodes.");
        }

        // vector of RecordTokens each representing a homolog
        Vector vecTokens = new Vector();

        // loop over all homologs
        while((node = nodesHits.nextNode()) != null) {
            String strGI, strStart, strEnd, strQueryStart;
            try {
                strGI = XPathAPI.selectSingleNode(node, p_xpath_gi + "/child::text()").getNodeValue();
                strStart = XPathAPI.selectSingleNode(node, p_xpath_start + "/child::text()").getNodeValue();
                strEnd = XPathAPI.selectSingleNode(node, p_xpath_end + "/child::text()").getNodeValue();
                strQueryStart = XPathAPI.selectSingleNode(node, p_xpath_query_start + "/child::text()").getNodeValue();
            } catch(javax.xml.transform.TransformerException e) {
                System.err.println("EnumHomologs: could not select node.");
                e.printStackTrace();
                throw new IllegalActionException("EnumHomologs: could not select node.");
            }

            int intStart = Integer.parseInt(strStart);
            int intEnd = Integer.parseInt(strEnd);
            int intQueryStart = Integer.parseInt(strQueryStart);

            //gi|1787566|gb|AE000229.1|AE000229
            _debug("Entire GI string: " + strGI);
            // extract the GI from strGI
            strGI = strGI.split("\\|")[1];

            _debug("EnumHomologs: GI=" + strGI + "\tStart=" + intStart + "\tEnd=" + intEnd);


            // construct a RecordToken and add it to vector
            vecTokens.add(new RecordToken("{GeneID = '" + strGI +
                                        "', Start = " + intStart +
                                        ", End = " + intEnd +
                                        ", QueryStart = " + intQueryStart + "}"));
        }

        // construct an ArrayToken and broadcast that
        portOutput.broadcast(new ArrayToken((RecordToken[]) vecTokens.toArray(new RecordToken[0])));
    }
}

// vim: ts=4 sw=4 et
