/**
 *    '$RCSfile: DataTypeResolver.java,v $'
 *
 *     '$Author: rspears $'
 *       '$Date: 2004/09/20 15:02:29 $'
 *   '$Revision: 1.5 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.datasource;

import org.apache.xpath.XPathAPI;
import org.ecoinformatics.util.Config;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class takes in information from the metadata parsers and tries to
 * intelligently guess as to the data type of an attribute based on
 * provided metadata.  since different metadata standards provide differing
 * levels of complexity in their attribute level descriptions, some guesses
 * may be better than others, but this class, should at least be able
 * to distinguish between, whole (integer), real (float) and text (string)
 * types.
 */
public class DataTypeResolver
{
  private static DataTypeResolver resolver = null; //singleton variable
  private static String DEFAULT_TYPE = "STRING";

  private DataType[] dataTypes;

  /**
   * constructor
   */
  public DataTypeResolver()
  {
    try
    {
      NodeList nl = Config.getInstance().getDataTypes();
      dataTypes = new DataType[nl.getLength()];
      for(int i=0; i<nl.getLength(); i++)
      {
        Node n = nl.item(i);
        Node nameNode = XPathAPI.selectSingleNode(n, "./name");
        Node numberTypeNode = XPathAPI.selectSingleNode(n, "./numberType");
        if(nameNode == null || numberTypeNode == null)
        {
          throw new RuntimeException("Expecting a name and numberType node " +
            "child of dataType in " +
            "config.xml.");
        }

        String name = nameNode.getFirstChild().getNodeValue();
        String numberType = numberTypeNode.getFirstChild().getNodeValue();

        NodeList aliasNL = XPathAPI.selectNodeList(n, "./alias");
        String[] aliases = new String[aliasNL.getLength()];
        for(int j=0; j<aliasNL.getLength(); j++)
        {
          Node aliasNode = aliasNL.item(j);
          aliases[j] = aliasNode.getFirstChild().getNodeValue();
        }

        Node numericTypeNode = XPathAPI.selectSingleNode(n, "./numericType");
        String maxval = null;
        String minval = null;
        String textencoding = null;
        if(numericTypeNode != null)
        {
          Node minValNode = XPathAPI.selectSingleNode(numericTypeNode, "./minValue");
          Node maxValNode = XPathAPI.selectSingleNode(numericTypeNode, "./maxValue");
          if(minValNode == null || maxValNode == null)
          {
            throw new RuntimeException("Expecting a minValue and " +
              "maxValue children of numericType in config.xml.");
          }

          minval = minValNode.getFirstChild().getNodeValue();
          maxval = maxValNode.getFirstChild().getNodeValue();
        }
        else
        {
          Node textTypeNode = XPathAPI.selectSingleNode(n, "./textType");
          if(textTypeNode == null)
          {
            throw new RuntimeException("Each dataType defined in the " +
              "config.xml file must have either a textType or a numericType " +
              "defined to be valid.");
          }

          Node textEncodingNode = XPathAPI.selectSingleNode(textTypeNode, "./encoding");
          if(textEncodingNode == null)
          {
            throw new RuntimeException("The textType node must have an " +
              "encoding node as a child in config.xml.");
          }
          textencoding = textEncodingNode.getFirstChild().getNodeValue();
        }

        int numType = -1;
        if(numberType.equals("real"))
        {
          numType = DataType.REAL;
        }
        else if(numberType.equals("integer"))
        {
          numType = DataType.INTEGER;
        }
        else if(numberType.equals("nan"))
        {
          numType = DataType.STRING;
        }

        double min = -1.0;
        double max = -1.0;
        if(minval != null && maxval != null)
        {
          min = Double.parseDouble(minval);
          max = Double.parseDouble(maxval);
        }
        DataType dt = new DataType(name, min, max, numType, aliases);
        dataTypes[i] = dt;

      }
    }
    catch(javax.xml.transform.TransformerException te)
    {
      throw new RuntimeException("Error executing xpath expression in " +
        "DataTypeResolver.");
    }
    catch(Exception e)
    {
      throw new RuntimeException("Exception in DataTypeResolver.");
    }
  }

  /**
   * returns a singleton instance of the DataTypeResolver
   */
  public static DataTypeResolver instanceOf()
  {
    if(resolver == null)
    {
      resolver = new DataTypeResolver();
    }
    return resolver;
  }

  /**
   * returns an array of the data types defined in the config.xml file
   */
  public DataType[] getDataTypes()
  {
    return dataTypes;
  }

  /**
   * this method makes an educated guess as to the data type of an attribute.
   * min and max are Double so that they can be set to null.  If min and max
   * are null, then the method will attempt to match typeName with one of the
   * aliases in the config.xml file.  If min and max are not null, it will
   * first try to find a suitable range.  If no range is found, it will fall
   * back to the alias lookup.  If the method is sure that it cannot determine
   * the type, it will throw an InvalidTypeException.
   * @param min the minimum value for the attribute
   * @param max the maximum value for the attribute
   * @param typeName the name of the type declared in the metadata
   * @param numberType the type of number (REAL, INTEGER, STRING) taken from
   * the constants defined in DataType.java
   * @throws UnresolvableTypeException
   * @return one of the DataType constants (INT, LONG, FLOAT, DOUBLE, STRING)
   * declared in DataType.java
   */
  public String resolveType(Double min, Double max,
                            Integer numberType, String typeName)
    throws UnresolvableTypeException
  {
    if (typeName.equals("")) {
        typeName = DEFAULT_TYPE;
    }
    if(min == null || max == null)
    {
      return aliasLookup(typeName);
    }
    else if(min != null && max != null && numberType != null)
    {
      return rangeLookup(min.doubleValue(), max.doubleValue(),
                         numberType.intValue());
    }
    else
    {
      throw new UnresolvableTypeException("Unable to resolve data type with " +
        "parameters min: " + min + " max: " + max + " numberType: " +
        " typeName: " + typeName);
    }
  }

  /**
   * resolves the type based solely on the name.  This does an alias lookup
   * on the types in config.xml.
   * @param name the name of the type to resolve.
   */
  public String resolveType(String name)
    throws UnresolvableTypeException
  {
    return aliasLookup(name);
  }

  /**
   * return a DataType based on a range
   */
  private String rangeLookup(double min, double max, int numberType)
    throws UnresolvableTypeException
  {
    if(min > max)
    {
      throw new UnresolvableTypeException("The min cannot be greater than " +
        "the max.");
    }

    for(int i=0; i<dataTypes.length; i++)
    {
      double dtmin = dataTypes[i].getMin();
      double dtmax = dataTypes[i].getMax();

      if(dtmin < min && dtmax > max && numberType == dataTypes[i].getType())
      {
        return dataTypes[i].getName();
      }
    }

    throw new UnresolvableTypeException("The range lookup for the type with " +
      "parameters min: " + min + " max: " + max + " numberType: " + numberType +
      "failed.  No compatible types were found in the config.xml file.");
  }

  /**
   * find a DataType based on an alias
   */
  private String aliasLookup(String typeName)
    throws UnresolvableTypeException
  {
    for(int i=0; i<dataTypes.length; i++)
    {
      String aliases[] = dataTypes[i].getNames();
      String name = dataTypes[i].getName();
      for(int j=0; j<aliases.length; j++)
      {
        if(typeName.trim().equals(aliases[j].trim()) ||
           typeName.trim().equals(name))
        {
          return name;
        }
      }
    }

    throw new UnresolvableTypeException("The type '" + typeName.trim() +
      "' cannot be " +
      "resolved with any type in the config.xml file.  Please add an alias " +
      "for this type to one of the DataType elements in config.xml.");
  }
}
