/**
 *    '$RCSfile: DarwinCoreDataSource.java,v $'
 *
 *     '$Author: rspears $'
 *       '$Date: 2004/10/14 14:35:16 $'
 *   '$Revision: 1.4 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.datasource.darwincore;

import java.io.StringReader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import org.ecoinformatics.ecogrid.EcogridQueryParser;
import org.ecoinformatics.ecogrid.EcogridResultsetTransformer;
import org.ecoinformatics.ecogrid.QueryType;
import org.ecoinformatics.ecogrid.ResultsetType;
import org.ecoinformatics.ecogrid.ResultsetType_record;
import org.ecoinformatics.ecogrid.ResultsetType_record_returnField;
import org.ecoinformatics.ecogrid.digir.impl.DigirJavaToEcogridJavaResultsetTransformer;
import org.ecoinformatics.ecogrid.digir.impl.DigirProxyImpl;
import org.ecoinformatics.ecogrid.digir.impl.EcogridJavaToDigirJavaQueryTransformer;
import org.ecoinformatics.seek.dataquery.DBTablesGenerator;
import org.ecoinformatics.seek.datasource.DSSchemaDef;
import org.ecoinformatics.seek.datasource.DSSchemaIFace;
import org.ecoinformatics.seek.datasource.DSTableDef;
import org.ecoinformatics.seek.datasource.DSTableFieldIFace;
import org.ecoinformatics.seek.datasource.DSTableIFace;
import org.ecoinformatics.seek.datasource.DataCacheItem;
import org.ecoinformatics.seek.datasource.DataCacheListener;
import org.ecoinformatics.seek.datasource.DataCacheManager;
import org.ecoinformatics.seek.datasource.DataType;
import org.ecoinformatics.seek.datasource.EcogridDataCacheItem;
import org.ecoinformatics.seek.datasource.QBTableauFactory;
import org.ecoinformatics.seek.datasource.TableEntity;
import org.ecoinformatics.seek.ecogrid.ResultTreeRoot;
import org.ecoinformatics.seek.querybuilder.DBQueryDef;
import org.ecoinformatics.seek.querybuilder.DBQueryDefParserEmitter;
import org.ecoinformatics.seek.querybuilder.DBSchemaParserEmitter;

import ptolemy.actor.TypedIOPort;
import ptolemy.actor.lib.Source;
import ptolemy.data.DoubleToken;
import ptolemy.data.IntToken;
import ptolemy.data.LongToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.expr.FileParameter;
import ptolemy.data.type.BaseType;
import ptolemy.data.type.Type;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Settable;
import ptolemy.kernel.util.StringAttribute;
import ptolemy.vergil.icon.EditorIcon;
import ptolemy.vergil.kernel.attributes.ResizablePolygonAttribute;
import ptolemy.vergil.kernel.attributes.TextAttribute;

/**
 * A ResultRecord that is in EML 2.0.0 format.
 */
public class DarwinCoreDataSource extends Source implements DataCacheListener
{
    static final   String    DATATYPES[] = {DataType.INT, DataType.FLOAT, DataType.DOUBLE, DataType.LONG, DataType.STR};
    static final   BaseType  BASETYPES[] = {BaseType.INT, BaseType.DOUBLE, BaseType.DOUBLE, BaseType.LONG, BaseType.STRING};
    static private Hashtable mTypeHash   = new Hashtable();
   
    private org.ecoinformatics.util.DebugPrinter dbg;
          
    public  QBTableauFactory    _qbTableauFactory = null;
    
    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////

    /** Output indicator parameter.
     */
    private int                  _outputType         = _ASFIELD;
    private TableEntity          _tableEntity        = null;
    private Vector[]             _dataVectors        = null;
    private int                  _recordCount        = 0;
    
    private Vector               _columns            = null;
    private Vector               _colTypes           = new Vector();
        
    private DBQueryDef           _queryDef           = null;
    private DSSchemaIFace        _schemaDef          = null;
    private boolean              _ignoreSchemaChange = false;
    private boolean              _schemaAttrIsSet    = false;
    
    private DBTablesGenerator    _tableGenerator     = null;
    private EcogridDataCacheItem _cachedDataItem     = null;
    private ResultsetType        _resultsetType      = null;
    private ResultsetType_record[] _records          = null;

    /** Icon indicating the communication region. */
    protected ResizablePolygonAttribute _poly1;
    protected ResizablePolygonAttribute _poly2;
    protected TextAttribute             _text;
    
    
    // static members
    private static DSSchemaIFace _darwinCoreSchema   = null;
    
    // Constants used for more efficient execution.
    private static final int     _ASTABLE = 0;
    private static final int     _ASROW   = 1;
    private static final int     _ASFIELD = 2;
    
    protected static final String YELLOW       = "{1.0, 1.0, 0.0, 1.0}";
    protected static final String RED          = "{1.0, 0.0, 0.0, 1.0}";
    protected static final String BLACK        = "{0.0, 0.0, 0.0, 1.0}";
    protected static final String MAGENTA      = "{1.0, 0.0, 1.0, 1.0}";
    protected static final String TITLE_BINARY = "0101";
    protected static final String TITLE_BUSY   = "BUSY";
    protected static final String TITLE_ERROR  = "ERROR";
    
    
    ///////////////////////////////////////////////////////////////////
    ////                     ports and parameters                  ////
    
    public  FileParameter       _datasetPathAttr = null;
    public  StringAttribute     _sqlAttr         = null;
    public  StringAttribute     _schemaAttr      = null;
    public  StringAttribute     _outputTypeAttr  = null;
    public  StringAttribute     _endPoint        = null;
    public  StringAttribute     _provider        = null;
    public  StringAttribute     _searchData      = null;

    /**
     * Construct an actor with the given container and name.
     *
     * @param container The container.
     * @param name The name of this actor.
     * @exception IllegalActionException If the actor cannot be contained by the
     *                proposed container.
     * @exception NameDuplicationException If the container already has an actor
     *                with this name.
     * @since
     */
    public DarwinCoreDataSource(CompositeEntity container, String name)
                    throws NameDuplicationException, IllegalActionException
    {
        super(container, name);
        dbg = org.ecoinformatics.util.DebugPrinter.getInstance();
        //_attachText("_iconDescription", getReadyIcon());
        createIcon();
        
        _searchData = new StringAttribute(this, "searchData");
        _endPoint   = new StringAttribute(this, "endpoint");
        _provider   = new StringAttribute(this, "provider");
        
        // create tableau for editting the SQL String
       // _qbTableauFactory = new QBTableauFactory(this, "_tableauFactory");

        if (mTypeHash.size() == 0)
        {
            for (int i=0;i<DATATYPES.length;i++) 
            {
                mTypeHash.put(DATATYPES[i], BASETYPES[i]);       
            }
        }
        
        // Get the Schema for All DarwinCoreDataSources
        if (_darwinCoreSchema == null)
        {
            _darwinCoreSchema = DarwinCoreSchema.getInstance().getSchema();
        }
        
   }
/*    protected String getReadyIcon()
    {
        return  "<svg xmlns=\"http://www.w3.org/2000/svg\"" +
                "     width=\"22.438\" height=\"23.125\" viewBox=\"0 0 22.438 23.125\"" +
                "     style=\"overflow:visible;enable-background:new 0 0 22.438 23.125\"" +
                "     xml:space=\"preserve\">\n" +
                "    <polygon  style=\"fill:#F7F999;stroke:#D7AA2A;stroke-linejoin:round;\" " +
                "          points=\"6.5,-26.5 6.5,-7.875 27.917,-10.167 27.917,-27 16,-27 15,-28.563 7.063,-28.563\"/>\n" +
                "    <polyline style=\"fill:none;stroke:#D7AA2A;stroke-linejoin:round;\" " +
                "          points=\"6.5,-19.833 8.333,-22.5 14.417,-22.5 16.417,-24.5 20.417,-24.5 21.417,-25.625 27.917,-25.625\"/>\n" +
                "    <text x=\"10\" y=\"-12\" " +
                "          style=\"fill:#108531; font-family:'Courier-Bold'; font-size:8;\">0101</text>\n" +
                "</svg>\n";
    }*/

    /**
     * Create the folder icon
     * @throws NameDuplicationException
     * @throws IllegalActionException
     */
    private void createIcon()
        throws NameDuplicationException, IllegalActionException
    {
        EditorIcon node_icon = new EditorIcon(this, "_icon");

        // Create the folder
        _poly1 = new ResizablePolygonAttribute(node_icon, "outer");
        //_poly1.vertices.setToken("{7,-27, 7,-8, 28,-10, 28,-27, 16,-27, 15,-29, 7,-29}");
        _poly1.vertices.setToken("{0,0, 8,-2, 9,1, 24,-2, 24,20, 0,24, 0,0}");
        _poly1.width.setToken("24");
        _poly1.height.setToken("24");
        _poly1.centered.setToken("true");
        _poly1.lineColor.setToken(BLACK);

        _poly2 = new ResizablePolygonAttribute(node_icon, "inner");
        _poly2.vertices.setToken("{0,2, 24,-2, 24,20, 0,24, 0,2}");
        _poly2.width.setToken("24");
        _poly2.height.setToken("24");
        _poly2.centered.setToken("true");
        _poly2.lineColor.setToken(BLACK);
        
       
        _text = new TextAttribute(node_icon, "text");
        _text.textSize.setExpression("11");
        
        setIconStatus(TITLE_BINARY, YELLOW);

        node_icon.setPersistent(false); 
    }
    
    /**
     * Set the text and color of the icon
     * @param aText
     * @param aColor
     */
    protected void setIconStatus(String aText, String aColor)
    {
        try
        {
            _text.text.setExpression(aText);
            _poly1.fillColor.setToken(aColor);
            _poly2.fillColor.setToken(aColor);
        } catch (Exception e){}
    }

    /**
     * Generic StringAttribute Setter
     * 
     * @param aStrAttr the StringAttribute object being set
     * @param aVal the value being set
     * @param aName name of attr being set
     */
    protected void setStrAttr(StringAttribute aStrAttr, String aVal, String aName)
    {
        try {
            aStrAttr.setExpression(aVal);
        } catch (IllegalActionException iae) {
            System.err.println("Could not set "+aName+" with value[" + aVal + "]");
        }
    }

    /**
     * Generic StringAttribute Getter
     * @param aStrAttr the StringAttribute object being gotten
     * @param aName name of attr being gotten
     * 
     * @return value of attr
     */
    public String getStrAttr(StringAttribute aStrAttr, String aName)
    {
        String value = null;
        try {
          value = aStrAttr.getExpression();
        } catch (Exception e) 
        {
            System.err.println("Error getting "+aName+" attribute.");
        }
        return value;
    }
    
    
    /**
     * Set the endpoint of this record. The endpoint indicates where the service
     * generating the record can be accessed.
     * 
     * @param aVal the URL of the service that contains the record
     */
    public void setEndpoint(String aVal)
    {
        setStrAttr(_endPoint, aVal, "endPoint");
    }

    /**
     * Get the endpoint of this record. The endpoint indicates where the service
     * generating the record can be accessed.
     * 
     * @return endpoint the URL of the service that contains the record
     */
    public String getEndpoint()
    {

        return getStrAttr(_endPoint, "endPoint");
    }
    
    /**
     * Set the endpoint of this record. The endpoint indicates where the service
     * generating the record can be accessed.
     * 
     * @param aVal the URL of the service that contains the record
     */
    public void setProvider(String aVal)
    {
        setStrAttr(_provider, aVal, "provider");
    }

    /**
     * Get the endpoint of this record. The endpoint indicates where the service
     * generating the record can be accessed.
     * 
     * @return endpoint the URL of the service that contains the record
     */
    public String getProvider()
    {

        return getStrAttr(_provider, "provider");
    }
   
    /**
     * Set the searchData of this record. 
     * 
     * @param aVal the URL of the service that contains the record
     */
    public void setSearchData(String aVal)
    {
        setStrAttr(_searchData, aVal, "searchData");
    }

    /**
     * Get the searchData of this record. 
     * 
     * @return endpoint the URL of the service that contains the record
     */
    public String getSearchData()
    {
        System.out.println("getSearchData: "+ getStrAttr(_searchData, "searchData"));
        return getStrAttr(_searchData, "searchData");
    }
   
    /**
     * Initialize the actor prior to running in the workflow. This reads the metadata
     * and configures the ports.
     *
     * @throws IllegalActionException
     */
    public void initialize() throws IllegalActionException
    {

        _recordCount = 0;
    }

    /**
     * @return Description of the Returned Value
     * @exception IllegalActionException Description of Exception
     * @since
     */
    public boolean prefire() throws IllegalActionException
    {
        return super.prefire();
    }

    /**
     * Send an entire tableas a single token over the port on each fire event.
     *
     * @exception IllegalActionException If there is no director.
     */
    public void fireAsTable() throws IllegalActionException
    {
        TypedIOPort pp = (TypedIOPort) this.getPort("DataTable");
        pp.send(0, new StringToken(new String(_cachedDataItem.getData())));
        
        pp = (TypedIOPort) this.getPort("Delimiter");
        String delim =  _tableEntity.getDelimiter();    
        pp.send(0, new StringToken(delim));
        
        pp = (TypedIOPort) this.getPort("numColumns");
        pp.send(0, new IntToken(_tableEntity.getAttributes().length));
        
    }
    
    /**
     * Send a row as a single token over the port on each fire event.
     *
     * @exception IllegalActionException If there is no director.
     */
    public void fireAsRow() throws IllegalActionException
    {
        StringBuffer rowOfData = new StringBuffer();
        String delim = _tableEntity.getDelimiter();
        
        TypedIOPort port = (TypedIOPort) this.getPort("DataRow");
        if (port != null)
        {
            Enumeration dataEnum = _dataVectors[_recordCount].elements();
            int fieldCnt = 0;
            while (dataEnum.hasMoreElements())
            {
                String eleStr = (String)dataEnum.nextElement();
      
                if (fieldCnt > 0) 
                {
                    rowOfData.append(delim);
                }
                rowOfData.append(eleStr);
            }
            
            port.send(0, new StringToken(rowOfData.toString()));
            
            port = (TypedIOPort) this.getPort("Delimiter");
            port.send(0, new StringToken(delim));
            
            port = (TypedIOPort) this.getPort("numColumns");
            port.send(0, new IntToken(_tableEntity.getAttributes().length));
            
            _recordCount++;
        }
        
    }
    
    /**
     * Send a record's tokens over the ports on each fire event.
     *
     * @exception IllegalActionException If there is no director.
     */
    public void fireAsField() throws IllegalActionException
    {
        if (_records != null && _recordCount == _records.length) return;
        
        //dbg.print("Processing record: " + i, 2);
        Enumeration colEnum  = _columns.elements();
        Enumeration typeEnum = _colTypes.elements();
        int colInx = 0;
        while (colEnum.hasMoreElements() && typeEnum.hasMoreElements())
        {
            DSTableFieldIFace colDef   = (DSTableFieldIFace)colEnum.nextElement();
            Type              type     = (Type)typeEnum.nextElement();
            TypedIOPort       port     = (TypedIOPort) this.getPort(colDef.getName().trim());
            
            ResultsetType_record_returnField field = _records[_recordCount].getReturnField(colInx);
            String                           eleStr   = field.getContent();
  
            if (port != null)
            {
                Token val = null;
              
                //find the data type for each att
                if (type == BaseType.INT) {
                    try {
                        val = new IntToken(new Integer(eleStr).intValue());
                    } catch (Exception e) {
                        val = new IntToken(0);
                    }
                } else if (type == BaseType.DOUBLE) {
                    val = new DoubleToken(new Double(eleStr).doubleValue());
                                    
                } else if (type == BaseType.LONG) {
                    val = new LongToken(new Long(eleStr).longValue());
                                    
                } else if (type == BaseType.STRING) {
                    val = new StringToken(eleStr);
                  
                } else {
                    val = new IntToken(0);
                }
                //send the data on the port
                port.send(0, val);
            }
            colInx++;
        }
        _recordCount++;
    }
    
    /**
     * Send a record's tokens over the ports on each fire event.
     *
     * @exception IllegalActionException If there is no director.
     */
    public void fire() throws IllegalActionException
    {
        super.fire();
        switch (_outputType)
        {
            case _ASTABLE:
                fireAsTable();
                break;
                
            case _ASROW:
                fireAsRow();
                break;
                
            case _ASFIELD:
                fireAsField();
                break;
                
            default:
                throw new IllegalActionException(this,
                        "Unrecognized outputType: " + _outputType);
                         
        }
    }

    /**
     * 
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void initializePort(String aPortName, String aType, boolean aIsNew) 
                                throws ptolemy.kernel.util.IllegalActionException
    {
        // Create a new port for each Column in the resultset
        try {
            String columnName = aPortName.trim();
            TypedIOPort port;
            if (aIsNew) 
            {
                // Create a new typed port and add it to this container
                port = new TypedIOPort(this, columnName, false, true);
            } else 
            {
                // Set the type of the existing port
                port = (TypedIOPort)this.getPort(columnName);
            }
            port.setTypeEquals(getBaseType(aType));

        } catch (ptolemy.kernel.util.NameDuplicationException nde) 
        {
            throw new IllegalActionException(
                "One or more attributes has the " +
                "same name.  Please correct this and try again.");
        }
    }

    private boolean isAlreadyConfigured(int aType)
                           throws ptolemy.kernel.util.IllegalActionException
    {
        switch (_outputType)
        {
            case _ASTABLE:
                return getPort("DataTable") != null &&
                       getPort("Delimiter") != null &&
                       getPort("numColumns") != null;
                
            case _ASROW:
                return getPort("DataRow") != null &&
                       getPort("Delimiter") != null &&
                       getPort("numColumns") != null;
                
            case _ASFIELD:
                return false; // don't know
                
            default:
                throw new IllegalActionException(this,
                        "Unrecognized outputType: " + _outputType);
                         
        }        
    }
    
    /** Remove all ports by setting their container to null.
     *  As a side effect, the ports will be unlinked from all relations.
     *  This method is write-synchronized on the workspace, and increments
     *  its version number.
     */
    public void removeAllOutputPorts() 
      throws ptolemy.kernel.util.IllegalActionException
    {
        Vector portNames = getExistingPortNames();
        for (Enumeration e = portNames.elements(); e.hasMoreElements();)
        {
            String currPortName = (String)e.nextElement();
            TypedIOPort port = (TypedIOPort)this.getPort(currPortName);
            if (port != null && !port.isInput())
            {
                try 
                {
                  port.setContainer(null);
                } catch(Exception ex) 
                {
                    throw new IllegalActionException(this,
                            "Error removing port: " + currPortName);
                }
                  
            }
        }
        
    }
    
    /**
     * It reconfigures all the Ports to represent passing data back a row at a time
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void reconfigurePortsAsTable()
                        throws ptolemy.kernel.util.IllegalActionException
    {
        if (!isAlreadyConfigured(_outputType))
            removeAllOutputPorts();
        
        try {
            TypedIOPort pp = new TypedIOPort(this, "DataTable", false, true);
            pp.setTypeEquals(BaseType.STRING);
            
            pp = new TypedIOPort(this, "Delimiter", false, true);
            pp.setTypeEquals(BaseType.STRING);
            
            pp = new TypedIOPort(this, "numColumns", false, true);
            pp.setTypeEquals(BaseType.INT);
          } catch(Exception ex) 
          {
              throw new IllegalActionException(this,
                      "Error creating Table Ports ");
          }
    }
    
  
    /**
     * It reconfigures all the Ports to represent passing data back a row at a time
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void reconfigurePortsAsRow()
                        throws ptolemy.kernel.util.IllegalActionException
    {
        if (!isAlreadyConfigured(_outputType))
            removeAllOutputPorts();
        
        try {
            TypedIOPort pp = new TypedIOPort(this, "DataRow", false, true);
            pp.setTypeEquals(BaseType.STRING);
            
            pp = new TypedIOPort(this, "Delimiter", false, true);
            pp.setTypeEquals(BaseType.STRING);
            
            pp = new TypedIOPort(this, "numColumns", false, true);
            pp.setTypeEquals(BaseType.INT);
          } catch(Exception ex) 
          {
              throw new IllegalActionException(this,
                      "Error creating Row Ports ");
          }
    }
    
    /**
     * It reconfigures all the Ports from the new "columns" list
     * Meaning it removes port not in the list and
     * adds new port that are in the list but not yet there
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void reconfigurePortsAsField(Vector aNewColumns, Vector aColTypes)
                        throws ptolemy.kernel.util.IllegalActionException
    {
      // Get List of new column names
      Vector newColNames = new Vector();
      for (Enumeration e = aNewColumns.elements(); e.hasMoreElements();)
      {
          newColNames.addElement(((DSTableFieldIFace)e.nextElement()).getName());
      }
        
      // Get List of Existing Port names
      Vector currentNames = getExistingPortNames();
        
      // loop through the current column names and delete any "Current"
      // name that isn't in the list of "new" names
      for (Enumeration e = currentNames.elements(); e.hasMoreElements();)
      {
          String currPortName = (String)e.nextElement();
          TypedIOPort port = (TypedIOPort)this.getPort(currPortName);
          if (port != null && !newColNames.contains(currPortName) && !port.isInput()) 
          {
              //System.out.println("Remove port["+currPortName+"]"); // debug
              try {
                port.setContainer(null);
              } catch(Exception ex) 
              {
                  throw new IllegalActionException(this,
                          "Error removing port: " + currPortName);
              }
          }
      }
      
      // Now add all the new columns
      for (Enumeration e = aNewColumns.elements(); e.hasMoreElements();)
      {
          DSTableFieldIFace column = (DSTableFieldIFace)e.nextElement();
          String name = column.getName();
          boolean isNew = !currentNames.contains(name);
          initializePort(name, column.getDataType(), isNew);
      }           
      
      buildBaseTypeVectorFromColumns(aNewColumns, aColTypes);
    }

    /**
     * 
     * @param aNewColumns
     * @param aColTypes
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void reconfigurePorts(Vector aNewColumns, Vector aColTypes)
                        throws ptolemy.kernel.util.IllegalActionException
    {
       //if (isAlreadyConfigured(_outputType))
      //     return;
       
       switch (_outputType)
       {
           case _ASTABLE:
               reconfigurePortsAsTable();
               break;
               
           case _ASROW:
               reconfigurePortsAsRow();
               break;
               
           case _ASFIELD:
               reconfigurePortsAsField(aNewColumns, aColTypes);
               break;
               
           default:
               throw new IllegalActionException(this,
                       "Unrecognized outputType: " + _outputType);
                        
       }
    }

    
    /**
     * Callback for changes in attribute values.
     */
    public void attributeChanged(ptolemy.kernel.util.Attribute attribute)
                    throws ptolemy.kernel.util.IllegalActionException
    {
      //System.out.println("Changed: "+attribute.getName()+"  "+((Settable)attribute).getExpression());
      if (attribute == _sqlAttr) {
          String sqlDef = ((Settable)attribute).getDefaultExpression();
          sqlDef = ((Settable)attribute).getExpression();
          if (sqlDef.length() > 0)
          {
              _queryDef = DBQueryDefParserEmitter.parseQueryDef(_schemaDef, sqlDef);
              _columns  = _queryDef.getSelects();
              reconfigurePorts(_queryDef.getSelects(), _colTypes);
          }
        
      } else if (attribute == _schemaAttr && !_ignoreSchemaChange) // NOTE: We may skip setting it here because _ignoreSchemaChange may be true
      {
          String schemaDef = ((Settable)_schemaAttr).getExpression();
          
          // MOML may have a blank definition
          if (schemaDef.length() > 0)
          {
              _schemaAttrIsSet = true; // remember that we have been set by the MOML
              
              dbg.print("schemaDef >>"+schemaDef+"<<", 2);
              
              _schemaDef = DBSchemaParserEmitter.parseSchemaDef(schemaDef);
          }
                      
      } else if (attribute.getName().equals("provider") &&
                 !(attribute.getContainer().getContainer() instanceof ResultTreeRoot)) 
      {
          setIconStatus(TITLE_BUSY, RED);
          
          String endPointStr = getEndpoint();
          String providerStr = getProvider();
          String searchStr   = getSearchData();
          
          System.out.println("providerStr: "+providerStr);
          System.out.println("endPointStr: "+endPointStr);
          System.out.println("searchStr:   "+searchStr);
          
          StringBuffer queryBuf = new StringBuffer();
          queryBuf.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
          queryBuf.append("<egq:query queryId=\"query-digir.1.1\" system=\"http://knb.ecoinformatics.org\" ");
          queryBuf.append("  xmlns:egq=\"ecogrid://ecoinformatics.org/ecogrid-query-1.0.0beta1\" ");
          queryBuf.append("  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" ");
          queryBuf.append("  xsi:schemaLocation=\"ecogrid://ecoinformatics.org/ecogrid-query-1.0.0beta1 ../../src/xsd/query.xsd\">");
          queryBuf.append("  <namespace prefix=\"darwin\">http://digir.net/schema/conceptual/darwin/2003/1.0</namespace>");
          queryBuf.append("  <title>" + searchStr +" Query</title>");
          
          boolean useAllField = false;
          if (useAllField)
          {
              DSTableIFace table = (DSTableIFace)_darwinCoreSchema.getTables().elementAt(0); // there is only one table
              for (Enumeration e = table.getFields().elements(); e.hasMoreElements();)
              {
                  DSTableFieldIFace field = (DSTableFieldIFace)e.nextElement();
                  queryBuf.append("  <returnfield>/" + field.getName() + "</returnfield>");
              }
          } else 
          {
              queryBuf.append("  <returnfield>/ScientificName</returnfield>");
              queryBuf.append("  <returnfield>/Collector</returnfield>");
              queryBuf.append("  <returnfield>/YearCollected</returnfield>");
              queryBuf.append("  <returnfield>/InstitutionCode</returnfield>");
              queryBuf.append("  <returnfield>/CollectionCode</returnfield>");
              queryBuf.append("  <returnfield>/CatalogNumber</returnfield>");
              queryBuf.append("  <returnfield>/CatalogNumberText</returnfield>");
              queryBuf.append("  <returnfield>/Latitude</returnfield>");
              queryBuf.append("  <returnfield>/Longitude</returnfield>");
          }
          queryBuf.append("  <condition operator=\"LIKE\" concept=\"Genus\">" + searchStr +"</condition>");
          //queryBuf.append("  <condition operator=\"LIKE\" concept=\"Family\">canidae</condition>");
          queryBuf.append("</egq:query>");
          
          if (providerStr != null && providerStr.length() > 0)
          {
              try
              {
                  QueryType query = null;
                  
                  // XXX no error checking
                  int inx = providerStr.indexOf('?');
                  String baseURL = providerStr.substring(0, inx);
                  inx = providerStr.lastIndexOf('=');
                  String resource = providerStr.substring(inx+1, providerStr.length()).replaceAll("%20", " ");
                  //System.out.println(queryBuf.toString());
                  StringReader       strReader = new StringReader(queryBuf.toString());
                  EcogridQueryParser parser    = new EcogridQueryParser(strReader);
                  parser.parseXML();
                  query = parser.getEcogridQuery();
                  
                  EcogridJavaToDigirJavaQueryTransformer transformer = new EcogridJavaToDigirJavaQueryTransformer();
                  String digirQueryStr = transformer.transform(query, resource, baseURL);
                  System.out.println(digirQueryStr);
                  if (digirQueryStr.length() > 0)
                  {
                      setIconStatus(TITLE_BUSY, RED);
                      String resName = "ProviderData " + providerStr;
                      String fullURL = DigirProxyImpl.getFullDigirURL(digirQueryStr);
                      _cachedDataItem = (EcogridDataCacheItem)DataCacheManager.getInstance().getCacheItem(this, resName, fullURL, EcogridDataCacheItem.class.getName(), true);
                      _cachedDataItem.start();
                  }
              } catch (Exception e )
              {
                  System.err.println(e);
                  e.printStackTrace();
              }
          } else
          {
              setIconStatus(TITLE_ERROR, MAGENTA);
          }
      }
    }

    /**
     * returns the names of the current ports
     */
    private Vector getExistingPortNames()
    {
        List l = this.portList();
        Vector v = new Vector();
        for (int i = 0; i < l.size(); i++) {
            TypedIOPort p = (TypedIOPort) l.get(i);
            v.add(p.getName().trim());
        }
        return v;
    }
    
    /**
     * Returns a Ptolemly type for a given Kepler DataSource Type
     * @param aType DataSource type
     * @return Ptolemy type
     */
    private BaseType getBaseType(String aType)
    {
      BaseType type = (BaseType)mTypeHash.get(aType);
      if (type == null)
      {
          return BaseType.UNKNOWN;
      }
      return type;
    }

    /**
     * Fills the BaseType Vector with a BaseType for each column
     * @param aCols the columns of the DB
     * @param aBaseTypes the Vecotr to be filled
     */
    private void buildBaseTypeVectorFromColumns(Vector aCols, Vector aBaseTypes)
    {
        aBaseTypes.clear();
        for (Enumeration e = aCols.elements(); e.hasMoreElements();)
        {
            DSTableFieldIFace column = (DSTableFieldIFace)e.nextElement();
            aBaseTypes.addElement(getBaseType(column.getDataType()));
        }

    }
    
    //------------------------------------------------------------------------
    //-- DataCacheListener
    //------------------------------------------------------------------------
    
    public void complete(DataCacheItem aItem)
    {
        setIconStatus(TITLE_BINARY, YELLOW);
        if (_cachedDataItem.isReady()) 
        {
                  
            boolean resultsetWasOK = false;
            try 
            {
                if (_cachedDataItem.getData() != null)
                {
                    String resultsetStr = new String(_cachedDataItem.getData());
                    System.out.println(resultsetStr);
                    DigirJavaToEcogridJavaResultsetTransformer ecogrid2ResultTransformer = new DigirJavaToEcogridJavaResultsetTransformer();
                    _resultsetType = ecogrid2ResultTransformer.transform(resultsetStr);
                    if (_resultsetType != null)
                    {
                        System.out.println("Resultset was: "+EcogridResultsetTransformer.toXMLString(_resultsetType));
                    } else 
                    {
                        System.out.println("Resultset was NULL!");
                    }
                      
                    if (_resultsetType != null) 
                    {
                        _records = _resultsetType.getRecord();
                        if (_records != null)
                        {
                            resultsetWasOK = true;
                            _schemaDef = new DSSchemaDef();
                            DSTableDef table = new DSTableDef("DarwinCore");
                            ResultsetType_record_returnField[] fields = _records[0].getReturnField();
                            DarwinCoreSchema dcSchema = DarwinCoreSchema.getInstance();
                            for (int i=0;i<fields.length;i++) 
                            {
                                table.addField(fields[i].getName(), dcSchema.lookupTypeFromName(fields[i].getName()));
                            }
                            
                            try
                            {
                                //DSTableIFace table = (DSTableIFace)_darwinCoreSchema.getTables().elementAt(0);
                                _columns = table.getFields();
                                reconfigurePortsAsField(table.getFields(), _colTypes);
                            } catch (Exception e)
                            {
                                e.printStackTrace();
                            }
                        }
                    }
                }
            } catch (Exception e)
            {
                setIconStatus(TITLE_ERROR, MAGENTA);            
            }
              
            if (resultsetWasOK)
            {
                DataCacheManager.getInstance().saveCache();
            } else 
            {
                setIconStatus(TITLE_ERROR, MAGENTA);              
            }
        } else
        {
            setIconStatus(TITLE_ERROR, MAGENTA);            
        }

    }

}

