/**
 *    '$RCSfile: Eml200DataSource.java,v $'
 *
 *     '$Author: rspears $'
 *       '$Date: 2004/10/14 14:35:16 $'
 *   '$Revision: 1.28 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.datasource.eml.eml2;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import org.ecoinformatics.seek.dataquery.DBTableNameResolver;
import org.ecoinformatics.seek.dataquery.DBTablesGenerator;
import org.ecoinformatics.seek.dataquery.HsqlDataQueryAction;
import org.ecoinformatics.seek.datasource.DSSchemaDef;
import org.ecoinformatics.seek.datasource.DSSchemaIFace;
import org.ecoinformatics.seek.datasource.DSTableFieldIFace;
import org.ecoinformatics.seek.datasource.DataCacheItem;
import org.ecoinformatics.seek.datasource.DataCacheListener;
import org.ecoinformatics.seek.datasource.*;
import org.ecoinformatics.seek.datasource.DataType;
import org.ecoinformatics.seek.datasource.EcogridDataCacheItem;
import org.ecoinformatics.seek.datasource.EcogridMetaDataCacheItem;
import org.ecoinformatics.seek.datasource.QBTableauFactory;
import org.ecoinformatics.seek.datasource.TableEntity;
import org.ecoinformatics.seek.ecogrid.ResultRecord;
import org.ecoinformatics.seek.ecogrid.ResultTreeRoot;
import org.ecoinformatics.seek.querybuilder.DBQueryDef;
import org.ecoinformatics.seek.querybuilder.DBQueryDefParserEmitter;
import org.ecoinformatics.seek.querybuilder.DBSchemaParserEmitter;
import org.ecoinformatics.util.DelimitedReader;
import org.xml.sax.InputSource;

import ptolemy.actor.TypedIOPort;
import ptolemy.actor.gui.style.TextStyle;
import ptolemy.data.DoubleToken;
import ptolemy.data.IntToken;
import ptolemy.data.LongToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.expr.FileParameter;
import ptolemy.data.type.BaseType;
import ptolemy.data.type.Type;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Settable;
import ptolemy.kernel.util.StringAttribute;
import ptolemy.data.expr.StringParameter;

/**
 * A ResultRecord that is in EML 2.0.0 format.
 */
public class Eml200DataSource extends ResultRecord implements DataCacheListener
{
    static final   String    DATATYPES[] = {DataType.INT, DataType.FLOAT, DataType.DOUBLE, DataType.LONG, DataType.STR};
    static final   BaseType  BASETYPES[] = {BaseType.INT, BaseType.DOUBLE, BaseType.DOUBLE, BaseType.LONG, BaseType.STRING};
    static private Hashtable _TypeHash   = new Hashtable();
   
    private org.ecoinformatics.util.DebugPrinter dbg;
          
    public  QBTableauFactory    _qbTableauFactory = null;
    
    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////

    /** Output indicator parameter.
     */
    private int                  _outputType         = _NOPORTTYPE;
    private TableEntity          _tableEntity        = null;
    private Vector[]             _dataVectors        = null;
    private int                  _recordCount        = 0;
    
    private Vector               _columns            = null;
    private Vector               _colTypes           = new Vector();
        
    private DBQueryDef           _queryDef           = null;
    private DSSchemaIFace        _schemaDef          = null;
    private boolean              _ignoreSchemaChange = false;
    private boolean              _schemaAttrIsSet    = false;
    
    private DBTablesGenerator        _tableGenerator = null;
    private EcogridDataCacheItem     _cachedDataItem = null;
    private EcogridMetaDataCacheItem _cacheMetaDataItem       = null;
    
    // Constants used for more efficient execution.
    private static final int     _NOPORTTYPE = 0;
    private static final int     _ASTABLE    = 1;
    private static final int     _ASROW      = 2;
    private static final int     _ASFIELD    = 3;
    
    ///////////////////////////////////////////////////////////////////
    ////                     ports and parameters                  ////
    
    public  FileParameter       _datasetPathAttr = null;
    public  StringAttribute     _sqlAttr         = null;
    public  StringAttribute     _schemaAttr      = null;
    public  StringParameter     _outputTypeAttr  = null;

    /**
     * Construct an actor with the given container and name.
     *
     * @param container The container.
     * @param name The name of this actor.
     * @exception IllegalActionException If the actor cannot be contained by the
     *                proposed container.
     * @exception NameDuplicationException If the container already has an actor
     *                with this name.
     * @since
     */
    public Eml200DataSource(CompositeEntity container, String name)
                    throws NameDuplicationException, IllegalActionException
    {
        super(container, name);
        dbg = org.ecoinformatics.util.DebugPrinter.getInstance();
        _datasetPathAttr = new FileParameter(this, "EML File");
         
        _schemaAttr   = new StringAttribute(this, "schemaDef");
        TextStyle schemaDefTS = new TextStyle(_schemaAttr, "schemaDef");

        _sqlAttr = new StringAttribute(this, "sqlDef");
        TextStyle sqlDefTS = new TextStyle(_sqlAttr, "sqlDef");

        _outputTypeAttr = new StringParameter(this, "outputType");
        _outputTypeAttr.setExpression("As Field");
        _outputTypeAttr.addChoice("As Field");
        _outputTypeAttr.addChoice("As Table");
        _outputTypeAttr.addChoice("As Row");
        _outputType = _ASFIELD;
        
        // create tableau for editting the SQL String
        _qbTableauFactory = new QBTableauFactory(this, "_tableauFactory");

        if (_TypeHash.size() == 0)
        {
            for (int i=0;i<DATATYPES.length;i++) 
            {
                _TypeHash.put(DATATYPES[i], BASETYPES[i]);       
            }
        }
    }

    /**
     * Initialize the actor prior to running in the workflow. This reads the metadata
     * and configures the ports.
     *
     * @throws IllegalActionException
     */
    public void initialize() throws IllegalActionException
    {
        
        if (_tableEntity == null)
        {
            throw new IllegalActionException("_tableEntity is NULL!");
        }
        
        String sqlStr = "";
        
        String sqlXMLStr = ((Settable)_sqlAttr).getExpression();
        if (sqlXMLStr == null || sqlXMLStr.length() == 0)
        {
            sqlStr = "SELECT * FROM " + (_tableEntity.getMappedName() != null ? _tableEntity.getMappedName() : _tableEntity.getName());
            
        } else 
        {
            Hashtable mappedNameHash = new Hashtable();
            /*         for (Enumeration e = _tableEntity.getColumns().elements(); e.hasMoreElements();)
                     {
                         Object obj = e.nextElement();
                         DSTableIFace table = (DSTableIFace)obj;
                         mappedNameHash.put(table.getName(), table.getMappedName());
                         System.out.println("Mapping ["+table.getName()+"]["+table.getMappedName()+"]");
                     }*/
           if (_tableEntity.getMappedName() != null)
           {
               mappedNameHash.put(_tableEntity.getName(), _tableEntity.getMappedName());
           }
           DBQueryDef queryDef = DBQueryDefParserEmitter.parseQueryDef(_schemaDef, sqlXMLStr, mappedNameHash);
           
           sqlStr = DBQueryDefParserEmitter.createSQL(_schemaDef, queryDef);
        }

         //System.out.println(sqlStr);

         // excuted query
        if (sqlStr != null && !sqlStr.trim().equals(""))
        {
            // if table gnerated successfully, we will run query
            if (_tableGenerator.getSuccessStatus())
            {
              try
              {
                setIconStatus(TITLE_BUSY, RED);
                HsqlDataQueryAction queryAction = new HsqlDataQueryAction();
                queryAction.setSQL(sqlStr);
                queryAction.actionPerformed(null);
                ResultSet result = queryAction.getResultSet();
                _dataVectors = transformResultSet(result);
                setIconStatus(TITLE_BINARY, YELLOW);
              }
              catch(Exception e)
              {
                dbg.print("Error to run query is "+e.getMessage(), 2);
                throw new IllegalActionException(e.getMessage());
              }

            }
         }
        _recordCount = 0;
    }
    
     /**
     * @return Description of the Returned Value
     * @exception IllegalActionException Description of Exception
     * @since
     */
    public boolean prefire() throws IllegalActionException
    {
        return super.prefire();
    }

    /**
     * Send an entire tableas a single token over the port on each fire event.
     *
     * @exception IllegalActionException If there is no director.
     */
    public void fireAsTable() throws IllegalActionException
    {
        TypedIOPort pp = (TypedIOPort) this.getPort("DataTable");
        pp.send(0, new StringToken(new String(_cachedDataItem.getData())));
        
        pp = (TypedIOPort) this.getPort("Delimiter");
        String delim =  _tableEntity.getDelimiter();    
        pp.send(0, new StringToken(delim));
        
        pp = (TypedIOPort) this.getPort("numColumns");
        pp.send(0, new IntToken(_tableEntity.getAttributes().length));
        
    }
    
    /**
     * Send a row as a single token over the port on each fire event.
     *
     * @exception IllegalActionException If there is no director.
     */
    public void fireAsRow() throws IllegalActionException
    {
        StringBuffer rowOfData = new StringBuffer();
        String delim = _tableEntity.getDelimiter();
        
        TypedIOPort port = (TypedIOPort) this.getPort("DataRow");
        if (port != null)
        {
            Enumeration dataEnum = _dataVectors[_recordCount].elements();
            int fieldCnt = 0;
            while (dataEnum.hasMoreElements())
            {
                String eleStr = (String)dataEnum.nextElement();
      
                if (fieldCnt > 0) 
                {
                    rowOfData.append(delim);
                }
                rowOfData.append(eleStr);
            }
            
            port.send(0, new StringToken(rowOfData.toString()));
            
            port = (TypedIOPort) this.getPort("Delimiter");
            port.send(0, new StringToken(delim));
            
            port = (TypedIOPort) this.getPort("numColumns");
            port.send(0, new IntToken(_tableEntity.getAttributes().length));
            
            _recordCount++;
        }
        
    }
    
    /**
     * Send a record's tokens over the ports on each fire event.
     *
     * @exception IllegalActionException If there is no director.
     */
    public void fireAsField() throws IllegalActionException
    {
        //dbg.print("Processing record: " + i, 2);
        Enumeration colEnum  = _columns.elements();
        Enumeration dataEnum = _dataVectors[_recordCount].elements();
        Enumeration typeEnum = _colTypes.elements();
        while (colEnum.hasMoreElements() && dataEnum.hasMoreElements() && typeEnum.hasMoreElements())
        {
            DSTableFieldIFace colDef   = (DSTableFieldIFace)colEnum.nextElement();
            String            eleStr   = (String)dataEnum.nextElement();
            Type              type     = (Type)typeEnum.nextElement();
            TypedIOPort       port     = (TypedIOPort) this.getPort(colDef.getName().trim());
  
            if (port != null)
            {
                Token val = null;
              
                //find the data type for each att
                if (type == BaseType.INT) {
                    try {
                        val = new IntToken(new Integer(eleStr).intValue());
                    } catch (Exception e) {
                        val = new IntToken(0);
                    }
                } else if (type == BaseType.DOUBLE) {
                    val = new DoubleToken(new Double(eleStr).doubleValue());
                                    
                } else if (type == BaseType.LONG) {
                    val = new LongToken(new Long(eleStr).longValue());
                                    
                } else if (type == BaseType.STRING) {
                    val = new StringToken(eleStr);
                  
                } else {
                    val = new IntToken(0);
                }
                //send the data on the port
                port.send(0, val);
            }
        }
        _recordCount++;
    }
    
    /**
     * Send a record's tokens over the ports on each fire event.
     *
     * @exception IllegalActionException If there is no director.
     */
    public void fire() throws IllegalActionException
    {
        super.fire();
        switch (_outputType)
        {
            case _ASTABLE:
                fireAsTable();
                break;
                
            case _ASROW:
                fireAsRow();
                break;
                
            case _ASFIELD:
                fireAsField();
                break;
                
            default:
                throw new IllegalActionException(this,
                        "Unrecognized outputType: " + _outputType);
                         
        }
    }

    /**
     * 
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void initializePort(String aPortName, String aType, boolean aIsNew) 
                                throws ptolemy.kernel.util.IllegalActionException
    {
        // Create a new port for each Column in the resultset
        try {
            String columnName = aPortName.trim();
            TypedIOPort port;
            if (aIsNew) 
            {
                // Create a new typed port and add it to this container
                port = new TypedIOPort(this, columnName, false, true);
                System.out.println("Creating port ["+columnName+"]"+this);
            } else 
            {
                // Set the type of the existing port
                port = (TypedIOPort)this.getPort(columnName);
                System.out.println("Existing port port ["+columnName+"]"+this);
            }
            port.setTypeEquals(getBaseType(aType));

        } catch (ptolemy.kernel.util.NameDuplicationException nde) 
        {
            throw new IllegalActionException(
                "One or more attributes has the " +
                "same name.  Please correct this and try again.");
        }
    }

    private boolean isAlreadyConfigured(int aType)
                           throws ptolemy.kernel.util.IllegalActionException
    {
        switch (_outputType)
        {
            case _ASTABLE:
                return getPort("DataTable") != null &&
                       getPort("Delimiter") != null &&
                       getPort("numColumns") != null;
                
            case _ASROW:
                return getPort("DataRow") != null &&
                       getPort("Delimiter") != null &&
                       getPort("numColumns") != null;
                
            case _ASFIELD:
                return false; // don't know
                
            default:
                throw new IllegalActionException(this,
                        "Unrecognized outputType: " + _outputType);
                         
        }        
    }
    
    /** Remove all ports by setting their container to null.
     *  As a side effect, the ports will be unlinked from all relations.
     *  This method is write-synchronized on the workspace, and increments
     *  its version number.
     */
    public void removeAllOutputPorts() 
      throws ptolemy.kernel.util.IllegalActionException
    {
        Vector portNames = getExistingPortNames();
        for (Enumeration e = portNames.elements(); e.hasMoreElements();)
        {
            String currPortName = (String)e.nextElement();
            TypedIOPort port = (TypedIOPort)this.getPort(currPortName);
            if (port != null && !port.isInput())
            {
                try 
                {
                  port.setContainer(null);
                } catch(Exception ex) 
                {
                    throw new IllegalActionException(this,
                            "Error removing port: " + currPortName);
                }
                  
            }
        }
        
    }
    
    /**
     * It reconfigures all the Ports to represent passing data back a row at a time
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void reconfigurePortsAsTable()
                        throws ptolemy.kernel.util.IllegalActionException
    {
        if (!isAlreadyConfigured(_outputType))
            removeAllOutputPorts();
        
        try {
            TypedIOPort pp = new TypedIOPort(this, "DataTable", false, true);
            pp.setTypeEquals(BaseType.STRING);
            
            pp = new TypedIOPort(this, "Delimiter", false, true);
            pp.setTypeEquals(BaseType.STRING);
            
            pp = new TypedIOPort(this, "numColumns", false, true);
            pp.setTypeEquals(BaseType.INT);
          } catch(Exception ex) 
          {
              throw new IllegalActionException(this,
                      "Error creating Table Ports ");
          }
    }
    
  
    /**
     * It reconfigures all the Ports to represent passing data back a row at a time
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void reconfigurePortsAsRow()
                        throws ptolemy.kernel.util.IllegalActionException
    {
        if (!isAlreadyConfigured(_outputType))
            removeAllOutputPorts();
        
        try {
            TypedIOPort pp = new TypedIOPort(this, "DataRow", false, true);
            pp.setTypeEquals(BaseType.STRING);
            
            pp = new TypedIOPort(this, "Delimiter", false, true);
            pp.setTypeEquals(BaseType.STRING);
            
            pp = new TypedIOPort(this, "numColumns", false, true);
            pp.setTypeEquals(BaseType.INT);
          } catch(Exception ex) 
          {
              throw new IllegalActionException(this,
                      "Error creating Row Ports ");
          }
    }
    
    /**
     * It reconfigures all the Ports from the new "columns" list
     * Meaning it removes port not in the list and
     * adds new port that are in the list but not yet there
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void reconfigurePortsAsField(Vector aNewColumns, Vector aColTypes)
                        throws ptolemy.kernel.util.IllegalActionException
    {
      // Get List of new column names
      Vector newColNames = new Vector();
      for (Enumeration e = aNewColumns.elements(); e.hasMoreElements();)
      {
          newColNames.addElement(((DSTableFieldIFace)e.nextElement()).getName());
      }
        
      // Get List of Existing Port names
      Vector currentNames = getExistingPortNames();
        
      // loop through the current column names and delete any "Current"
      // name that isn't in the list of "new" names
      for (Enumeration e = currentNames.elements(); e.hasMoreElements();)
      {
          String currPortName = (String)e.nextElement();
          TypedIOPort port = (TypedIOPort)this.getPort(currPortName);
          if (port != null && !newColNames.contains(currPortName) && !port.isInput()) 
          {
              //System.out.println("Remove port["+currPortName+"]"); // debug
              try {
                port.setContainer(null);
              } catch(Exception ex) 
              {
                  throw new IllegalActionException(this,
                          "Error removing port: " + currPortName);
              }
          }
      }
      
      // Now add all the new columns
      for (Enumeration e = aNewColumns.elements(); e.hasMoreElements();)
      {
          DSTableFieldIFace column = (DSTableFieldIFace)e.nextElement();
          String name = column.getName();
          boolean isNew = !currentNames.contains(name);
          initializePort(name, column.getDataType(), isNew);
      }           
      
      buildBaseTypeVectorFromColumns(aNewColumns, aColTypes);
      
      // Set color to yellow.
      setIconStatus(TITLE_BINARY, YELLOW);
    }

    /**
     * 
     * @param aNewColumns
     * @param aColTypes
     * @throws ptolemy.kernel.util.IllegalActionException
     */
    private void reconfigurePorts(Vector aNewColumns, Vector aColTypes)
                        throws ptolemy.kernel.util.IllegalActionException
    {
       setIconStatus(TITLE_BUSY, RED);
       
       switch (_outputType)
       {
           case _ASTABLE:
               reconfigurePortsAsTable();
               break;
               
           case _ASROW:
               reconfigurePortsAsRow();
               break;
               
           case _ASFIELD:
               reconfigurePortsAsField(aNewColumns, aColTypes);
               break;
               
           default:
               throw new IllegalActionException(this,
                       "Unrecognized outputType: " + _outputType);
                        
       }
       
       setIconStatus(TITLE_BINARY, YELLOW);
    }

    
    /**
     * Callback for changes in attribute values.
     */
    public void attributeChanged(ptolemy.kernel.util.Attribute attribute)
                    throws ptolemy.kernel.util.IllegalActionException
    {
     
      if (attribute == _sqlAttr) {
          String sqlDef = ((Settable)attribute).getDefaultExpression();
          sqlDef = ((Settable)attribute).getExpression();
          if (sqlDef.length() > 0)
          {
              _queryDef = DBQueryDefParserEmitter.parseQueryDef(_schemaDef, sqlDef);
              _columns  = _queryDef.getSelects();
              reconfigurePorts(_queryDef.getSelects(), _colTypes);
          }
        
      } else if (attribute == _schemaAttr && !_ignoreSchemaChange) // NOTE: We may skip setting it here because _ignoreSchemaChange may be true
      {
          String schemaDef = ((Settable)_schemaAttr).getExpression();
          
          // MOML may have a blank definition
          if (schemaDef.length() > 0)
          {
              _schemaAttrIsSet = true; // remember that we have been set by the MOML
              
              dbg.print("schemaDef >>"+schemaDef+"<<", 2);
              
              _schemaDef = DBSchemaParserEmitter.parseSchemaDef(schemaDef);
          }
                      
      } else if ((attribute.getName().equals("recordId") ||
                  attribute.getName().equals("endpoint")) &&
                 this.hasConnectionValues() &&
                 !(attribute.getContainer().getContainer() instanceof ResultTreeRoot)) 
      {
          setIconStatus(TITLE_BUSY, RED);
          
          _cacheMetaDataItem = (EcogridMetaDataCacheItem)DataCacheManager.getInstance().getCacheItem(this, "MetaData "+getRecordId(), getEndpoint(), EcogridMetaDataCacheItem.class.getName());
          _cacheMetaDataItem.setEndPoint(getEndpoint());
          _cacheMetaDataItem.setRecordId(getRecordId());
          _cacheMetaDataItem.start();
          
      } else if (attribute == _outputTypeAttr) {
          String strOutputType = _outputTypeAttr.stringValue();
          if (strOutputType.equals("As Table") && _outputType != _ASTABLE) {
              _outputType = _ASTABLE;
              reconfigurePortsAsTable();
              
          } else if (strOutputType.equals("As Row") && _outputType != _ASROW) {
              _outputType = _ASROW;
              reconfigurePortsAsRow();
              
          } else if (strOutputType.equals("As Field") && _outputType != _ASFIELD) {
              _outputType = _ASFIELD;
              if (_columns != null && _columns.size() > 0 &&
                  _colTypes != null && _colTypes.size() > 0)
                  {
                    reconfigurePortsAsField(_columns, _colTypes);
                  }
          } else {
              //throw new IllegalActionException(this,
              //        "Unrecognized outputType: " + strOutputType);
          }
          
      }
    }

    /**
     * parses the package and fills in the names and types arrays
     */
    private void parsePackage(String packageFile) throws IllegalActionException
    {
        File f = new File(packageFile);

        try { //parse the package for the names and types of the atts
            FileInputStream fis = new FileInputStream(f);
            parsePackage(fis);
        } catch (FileNotFoundException fnfe) {
            throw new IllegalActionException(
                            "Could not find EML file to parse. "
                                            + fnfe.getMessage());
        }
    }
    
    /**
     * Creates the schema definition from the incoming data columns
     */
    private void createSchemaFromData() 
    {
        try  // XXX shouldn't be catch this exception here
        {
        
            // _schemaAttrIsSet gets set when the schema has come from the MOML
            // So if it is false here, then we set the attribute from the data.
            //
            // If after this the attr gets set from the MOML it will override
            // what we set here.
            if (!_schemaAttrIsSet && _tableEntity != null)
            {
                DSSchemaDef schema = new DSSchemaDef();
                schema.addTable(_tableEntity);
                _schemaDef = schema;
                
                DBTableNameResolver nameResolver = new DBTableNameResolver();
                try 
                {
                  _tableEntity = nameResolver.resolveTableName(_tableEntity);
                } catch (Exception e){}
                
                _tableGenerator = new DBTablesGenerator(_tableEntity, _cachedDataItem.getFileName());
                _tableGenerator.run(); // don't do thread
                
                String schemaDefXML = DBSchemaParserEmitter.emitXML(_schemaDef);
                _ignoreSchemaChange = true;
                ((Settable)_schemaAttr).setExpression(schemaDefXML);
                _ignoreSchemaChange = false;
            }
        } catch (IllegalActionException e) 
        {
            dbg.print("In createSchemaFromData: " + e, 2);
        }

    }

    /**
     * parses the package and fills in the names and types arrays
     */
    private void parsePackage(InputStream eml) throws IllegalActionException
    {
        Eml200Parser eml2parser;

        try { //parse the package for the names and types of the atts
            dbg.print("creating parser", 2);
            eml2parser = new Eml200Parser();
            dbg.print("parsing...", 2);
            eml2parser.parse(eml);
            dbg.print("Done parsing", 2);
        } catch (Exception e) {
            throw new IllegalActionException("Error parsing the eml package: "
                            + e.getMessage());
        }
        _tableEntity  = getData(eml2parser); // fills in the _dataVectors data member
    }

    /**
     * parses the package and fills in the names and types arrays
     */
    private void parsePackage(Reader eml) throws IllegalActionException
    {
        Eml200Parser eml2parser;

        try { //parse the package for the names and types of the atts
            dbg.print("creating parser", 2);
            eml2parser = new Eml200Parser();
            dbg.print("parsing...", 2);
            eml2parser.parse(new InputSource(eml));
            dbg.print("Done parsing", 2);
        } catch (Exception e) {
            throw new IllegalActionException("Error parsing the eml package: "
                            + e.getMessage());
        }

        _tableEntity  = getData(eml2parser); // fills in the _dataVectors data member
    }

    /**
     * get the data based on the contents of the package
     *
     * @param parser the parser that has already parsed the package.
     */
    private TableEntity getData(Eml200Parser parser) throws IllegalActionException
    {
        
        if (parser.getEntityCount() > 1) {
            throw new IllegalActionException(
                "Currently this parser only deals " +
                "with one entity.  Please use a package with only one entity.");
        } else if (parser.getEntityCount() == 0) { 
            throw new IllegalActionException(
                "There must be at least one entity in the EML package."); 
        }

        Hashtable entityhash = parser.getEntityHash();
        TableEntity tableEntity = (TableEntity) entityhash.get(entityhash.keys().nextElement());
        dbg.print("entityhash: " + entityhash.toString(), 2);
        
        Hashtable filehash = parser.getDataFilesHash();
        dbg.print("filehash: " + filehash.toString(), 2);
        
        String fileURLStr = (String) filehash.get(filehash.keys().nextElement());
        dbg.print("Data URL is: " + fileURLStr, 2);

        setIconStatus(TITLE_BUSY, RED);
        
        _cachedDataItem = (EcogridDataCacheItem)DataCacheManager.getInstance().getCacheItem(this, "Data", fileURLStr, EcogridDataCacheItem.class.getName());
        _cachedDataItem.setEndPoint(this.getEndpoint());
        _cachedDataItem.start();
            
        return tableEntity;
    }
    
    /**
     * Creates the data member _dataVectors from a string buffer of data
     * @param aDataStr
     */
    protected void createDataVector(StringBuffer aDataStr)
    {
        String data = aDataStr.toString();
        System.out.println("-----------------\n"+data+"\n-----------------\n");
        
        //dbg.print("entityhash: " + entityhash.toString(), 2);
        DelimitedReader dr = new DelimitedReader(data,
                        _tableEntity.getAttributes().length, _tableEntity.getDelimiter(),
                        _tableEntity.getNumHeaderLines(), _tableEntity.getRecordDelimiter(), 
                        _tableEntity.getNumRecords());

        _dataVectors = dr.getTokenizedData(true);
    }
    
    /**
     * returns the names of the current ports
     */
    private Vector getExistingPortNames()
    {
        List l = this.portList();
        Vector v = new Vector();
        for (int i = 0; i < l.size(); i++) {
            TypedIOPort p = (TypedIOPort) l.get(i);
            v.add(p.getName().trim());
        }
        return v;
    }
    
    /**
     * Returns a Ptolemly type for a given Kepler DataSource Type
     * @param aType DataSource type
     * @return Ptolemy type
     */
    private BaseType getBaseType(String aType)
    {
      BaseType type = (BaseType)_TypeHash.get(aType);
      if (type == null)
      {
          return BaseType.UNKNOWN;
      }
      return type;
    }

    /**
     * Fills the BaseType Vector with a BaseType for each column
     * @param aCols the columns of the DB
     * @param aBaseTypes the Vecotr to be filled
     */
    private void buildBaseTypeVectorFromColumns(Vector aCols, Vector aBaseTypes)
    {
        aBaseTypes.clear();
        for (Enumeration e = aCols.elements(); e.hasMoreElements();)
        {
            DSTableFieldIFace column = (DSTableFieldIFace)e.nextElement();
            aBaseTypes.addElement(getBaseType(column.getDataType()));
        }

    }
    
    /*
     * Method to transform resultset to a vector array data. A vector is a row
     * and data type is string.
     */
    private Vector[] transformResultSet (ResultSet rs) throws SQLException
    {
      Vector [] dataArray = null;
      Vector data = new Vector();
      ResultSetMetaData metadata = rs.getMetaData();
      int columnSize  = metadata.getColumnCount();

      while (rs.next())
      {
        Vector rowData = new Vector();
        for (int i =0; i < columnSize; i++)
        {
          String str = rs.getString(i+1);
          rowData.add(str);
        }
        data.add(rowData);
      }
      dataArray = transformVectorToArray(data);
      return dataArray;
    }//transformREsultSet

    /**
     * 
     * @param vector
     * @return
     */
    private Vector[] transformVectorToArray(Vector vector)
    {
      if (vector == null)
      {
        return null;
      }
      int size = vector.size();
      Vector [] array = new Vector[size];
      for (int i=0 ; i < size; i++)
      {
        array[i]= (Vector) vector.elementAt(i);
      }
      return array;
    } //transformVectorToArray
    
    //------------------------------------------------------------------------
    //-- DataCacheListener
    //------------------------------------------------------------------------
    
    public void complete(DataCacheItem aItem)
    {
        System.out.println("complete: "+this);
        
        aItem.removeListener(this);
        
        if (aItem.isReady())
        {
            if (aItem instanceof EcogridDataCacheItem)
            {
                if (aItem.getData() != null)
                {
                    String dataStr = new String(aItem.getData());
                    StringBuffer dataSB = new StringBuffer(dataStr);
                    createDataVector(dataSB);
                    try
                    {
                      _columns = _tableEntity.getFields();
                      if (_outputType == _ASFIELD)
                      {
                        reconfigurePortsAsField(_columns, _colTypes);
                      }
                    } catch (Exception e){}
                    
                    createSchemaFromData();
                    setIconStatus(TITLE_BINARY, YELLOW);
                }
                
            } else if (aItem instanceof EcogridMetaDataCacheItem) {
                String data = new String(aItem.getData());
                try 
                {
                  StringReader recordReader = new StringReader(data);
                  parsePackage(recordReader);
                } catch (Exception e){}

                setIconStatus(TITLE_BINARY, YELLOW);
            }
            DataCacheManager.getInstance().saveCache();
        }
    }

}

