/**
 *    '$RCSfile: SimpleLibrarySearcher.java,v $'
 *
 *     '$Author: berkley $'
 *       '$Date: 2004/10/13 21:55:10 $'
 *   '$Revision: 1.8 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 *    '$RCSfile: SimpleLibrarySearcher.java,v $'
 *
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.ecogrid;

import javax.swing.JTree;
import javax.swing.tree.TreeModel;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import java.util.Stack;
import java.util.Vector;
import java.util.Iterator;

import ptolemy.moml.ErrorHandler;
import ptolemy.moml.MoMLParser;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.Entity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.NamedObj;

/**
 * class to search the library.  This class uses a simple name comparator
 * to determine results.  It uses a depth first traversal of the tree.
 */
public class SimpleLibrarySearcher extends LibrarySearcher
{
  private Stack pathStack; //the stack to create the return paths with

  /**
   * constructor
   */
  public SimpleLibrarySearcher(JTree library, LibrarySearchPane searchPane)
  {
    super(library, searchPane);
  }

  /**
   * search for val in the library
   */
  public LibrarySearchResults search(String val)
  {
    results = new LibrarySearchResults();
    if(val.trim().equals(""))
    {
      return results;
    }
    
    ErrorHandler handler = ptolemy.moml.MoMLParser.getErrorHandler();
    //temporarily disable error handling
    MoMLParser.setErrorHandler(new ErrorSkipper());
    TreeModel model = library.getModel();
    findval(val, model); //create the results vector

    //reinstate the rightful errorhandler
    MoMLParser.setErrorHandler(handler);
    return results;
  }

  /**
   * pre-reads the library so searches will be faster.
   * When the momlparser hits an error in the moml, it rebuilds the model
   * which in turn collapses the entire library JTree.  This
   * presearches the tree which causes the parse errors to get thrown before
   * the first search so the tree won't be collapsed the first time you
   * search.
   */
  protected void init()
  {
    pathStack = new Stack();
    //search("this is something that will not be found");
  }

  /**
   * look for the val in the TreeModel
   */
  private void findval(String val, TreeModel model)
  {
    Object o = model.getRoot(); //start from the root
    pathStack = new Stack();
    pathStack.push(o);
    for(int i=0; i<model.getChildCount(o); i++)
    {
      Object child = model.getChild(o, i);
      pathStack.push(child);
      String name = ((CompositeEntity)child).getName();
      compareVals(name, val);
      //go to the other findval method
      findval(val, (CompositeEntity)child);
    }
  }

  /**
   * look for val recursively inside entity
   */
  private void findval(String val, CompositeEntity entity)
  {
    for(Iterator i = entity.entityList().iterator(); i.hasNext();)
    {
      Entity e = (Entity)i.next();
      String name = e.getName();
      pathStack.push(e); //push our new candidate onto the stack

      if(e instanceof CompositeEntity)
      { //recurse into the tree
        compareVals(name, val);
        findval(val, (CompositeEntity)e); //keep searching
      }
      else
      { //leaf node
        compareVals(name, val);

        if(!pathStack.empty())
        { //remove the leaf node we just looked at
          pathStack.pop();
        }
      }
    }

    if(!pathStack.empty())
    { //remove the composite node we are currently searching
      pathStack.pop();
    }
  }

  /**
   * compares two strings and adds the path to the results if they
   * are a match
   */
  private void compareVals(String val1, String val2)
  {
    val1 = val1.toLowerCase(); //make the comparison case insensitive
    val2 = val2.toLowerCase();

    if(val1.indexOf(val2) != -1)
    {
      addStackTopToResult();
    }
  }

  /**
   * adds the path at the top of the stack to the result
   */
  private void addStackTopToResult()
  {
    results.add(createPathFromStack());
  }

  /**
   * return the path as a TreePath
   */
  private TreePath createPathFromStack()
  {
    return new TreePath(pathStack.toArray());
  }

  /**
   * errorhandler that ignore all errors and continues parsing.
   */
  private class ErrorSkipper implements ErrorHandler
  {
    public void enableErrorSkipping(boolean b)
    {//yeah, we're skipping everything whether you tell us to or not
    }

    public int handleError(java.lang.String element, NamedObj context,
      java.lang.Throwable exception)
    {
      return ErrorHandler.CONTINUE; //just keep it moving
    }
  }

  /**
   * A factory that creates the searcher to search the library
   */
  public static class Factory extends LibrarySearcherFactory
  {
    /** Create an factory with the given name and container.
     *  @param container The container.
     *  @param name The name of the entity.
     *  @exception IllegalActionException If the container is incompatible
     *   with this attribute.
     *  @exception NameDuplicationException If the name coincides with
     *   an attribute already in the container.
     */
    public Factory(NamedObj container, String name)
            throws IllegalActionException, NameDuplicationException
    {
      super(container, name);
    }

    /**
     *  creates a ScrollBarModifier and returns it.
     *  @param _libraryModel the model containing the actor library
     *  @return A new LibraryPane that displays the library
     */
    public LibrarySearcher createLibrarySearcher(JTree library, LibrarySearchPane searchPane)
    {
      return new SimpleLibrarySearcher(library, searchPane);
    }
  }
}
