/**
 *    '$RCSfile: DelimitedReader.java,v $'
 *
 *     '$Author: rspears $'
 *       '$Date: 2004/09/20 15:05:50 $'
 *   '$Revision: 1.9 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.util;

import java.io.*;
import java.util.*;

/**
 * tokenizes a delimited file.  This reader assumes that one record is on one
 * line which ends with the line
 */
public class DelimitedReader
{
  private String data;
  private Vector[] lines;
  private Vector linesVector;
  private int numHeaderLines;
  private int numRecords;
  private boolean stripHeader = false;
  private int currentRecord = 0;
  private int numCols;
  private String delimiter;
  private String lineEnding;
  private org.ecoinformatics.util.DebugPrinter dbg;

  /**
   * constructor. reads the csv stream.
   * @param delimString the delimited stream to read
   * @param numCols the number of columns in the stream
   * @param delimiter the delimiter to tokenize on
   * @param numHeaderLines the number of lines to skip at the top of the file
   * @param lineEnding the line ending char(s)...either "\n" (unix),
   * "\r\n" (windoze) or "\r" (mac)
   */
  public DelimitedReader(String data, int numCols, String delimiter,
    int numHeaderLines, String lineEnding, int numRecords)
  {
    dbg = org.ecoinformatics.util.DebugPrinter.getInstance();
    this.numHeaderLines = numHeaderLines;
    this.data = data;
    this.numCols = numCols;
    this.numRecords = numRecords;    
    dbg.print("Delimiter is: " + delimiter, 2);
    this.delimiter = unescapeDelimiter(delimiter);
    dbg.print("LineEnding is: " + lineEnding, 2);
    this.lineEnding = unescapeDelimiter(lineEnding);
    
    //lines = new Vector[numRecords + numHeaderLines + 1];
    linesVector = new Vector();

    int begin = 0;
    int end = 0;
//    int i = 0;
    while(end < data.length())
    { //add each line of the string as an element in a vector
      end = data.indexOf(this.lineEnding, begin); //DFH 'this.' added
      if(end == -1)
      {
        end = data.length();
      }
      String line = data.substring(begin, end);
      if(!line.trim().equals(""))
      {
        //take off the line ending
        // MBJ: I commented out the next line as it was improperly truncating lines
        // I'm not sure why it was there in the first place, as the previous substring
        // removed the delimiter
        //line = line.substring(0, line.length() - lineEnding.length());

        //split the line based on the delimiter
        String[] s = line.split(delimiter.trim(), numCols);
        Vector v = new Vector();
        for(int j=0; j<s.length; j++)
        {
          v.addElement(s[j]);
        }

        if(v.size()  < numCols)
        {
          int vsize = v.size();
          for(int j=0; j<numCols - vsize; j++)
          { //add any elements that aren't there so that all the records have the
            //same number of cols
            v.addElement("");
          }
        }
        //lines[i] = v;
        linesVector.add(v);
//        i++;
      }
      //go to the next line
      begin = end + this.lineEnding.length();   //DFH  'this.' added
    }
    
    int records = linesVector.size();
    if (records != this.numRecords) {
        this.numRecords = records;
        dbg.print("Metadata disagrees with actual data. Changing number of records to: " + records, 1);
    }
    lines = new Vector[records];
    for (int k=0; k < records; k++) {
        lines[k] = (Vector)linesVector.get(k);
    }
/*
    for(int j=0; j<lines.length; j++)
    {
      if(lines[j] == null)
      {
        lines[j] = new Vector();
      }
    }
*/
    
  }

  /**
   * returns the data as an array of vectors.  each vector will have the same
   * number of elements as there are columns in the data.
   * @param stripHeaderLines true if the header lines should not be included
   * in the returned data, false otherwise
   */
  public Vector[] getTokenizedData(boolean stripHeaderLines)
  {
    if(stripHeaderLines)
    {
      Vector[] strip = null;
      if (numRecords > numHeaderLines)
      {
        strip = new Vector[numRecords-numHeaderLines];
        for(int i=numHeaderLines; i<lines.length; i++)
        {
          strip[i-numHeaderLines] = lines[i];
        }
      }
      return strip;
    }
    else
    {
      return lines;
    }
  }

  /**
   * returns a string representation of the data
   */
  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    for(int i=0; i<lines.length; i++)
    {
      dbg.print("line[" + (i + 1) + "]: " + lines[i].toString(), 2);
      for(int j=0; j<lines[i].size(); j++)
      {
        sb.append((String)lines[i].elementAt(j));
        if(j != lines[i].size() - 1)
        {
          sb.append(" || ");
        }
      }
      sb.append(lineEnding);
    }
    return sb.toString();
  }
  
  /**
   * Convert a string escaped representation of a delimiter character into
   * an the actual String for that delimiter.  This is used for translating
   * escaped versions of tab, newline, and carriage return characters to
   * their real character values.
   * 
   * @param delimiter the String representing the delimiter
   * @return the actual String for the delimiter
   */
  private String unescapeDelimiter(String delimiter)
  {
      String newDelimiter = delimiter;
      
      if (delimiter.equals("\\t")) {
          dbg.print("Tab interpreted incorrectly as string.", 2);
          newDelimiter = "\t";
      } else if (delimiter.equals("\\n")) {
          dbg.print("Newline interpreted incorrectly as string.", 2);
          newDelimiter = "\n";
      } else if (delimiter.equals("\\r")) {
          dbg.print("CR interpreted incorrectly as string.", 2);
          newDelimiter = "\r";
      } else if (delimiter.equals("\\r\\n")) {
          dbg.print("CRNL interpreted incorrectly as string.", 2);
          newDelimiter = "\r\n";
      }
      return newDelimiter;
  }
}
