/** SSH2 actor to connect to remote hosts, execute commands, and get their
 *  output.
 *
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
package org.sdm.spa;

import ptolemy.actor.TypedAtomicActor;
import ptolemy.actor.TypedIOPort;
import ptolemy.actor.parameters.PortParameter;
import ptolemy.data.IntToken;
import ptolemy.data.StringToken;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Workspace;
import ptolemy.kernel.util.Attribute;

import com.jcraft.jsch.*;

import java.io.ByteArrayOutputStream;
import java.util.HashSet;

//////////////////////////////////////////////////////////////////////////
//// Ssh2Exec
/**
   Connects to a remote host using Ssh2 protocol.

   Error conditions this actor must respond robustly to:

   1. Wrong identity file given.
   2. Host unreachable.
   3. Login unsuccessful.
   4. Session dies prematurely.

   This actor will keep the session open until it receives a different
   username and host combination.

   Reference: Ant version ...
   @author Ilkay Altintas, Xiaowen Xin
   @version $Id: Ssh2Exec.java,v 1.1 2004/10/04 22:49:10 altintas Exp $
   @category.name remote
   @category.name connection
 */

public class Ssh2Exec extends TypedAtomicActor {


    /** Construct an SSH2 actor with the given container and name.
     *  Create the parameters, initialize their values.
     *  @param container The container.
     *  @param name The name of this actor.
     *  @exception IllegalActionException If the entity cannot be contained
     *  by the proposed container.
     *  @exception NameDuplicationException If the container already has an
     *  actor with this name.
     */
    public Ssh2Exec(CompositeEntity container, String name) throws
        NameDuplicationException, IllegalActionException {
        super(container, name);

        // initialize our variables
        _jsch = new JSch();
        _setIdentities = new HashSet();

        // create all the ports
        portUser = new PortParameter(this, "user");
        portHost = new PortParameter(this, "host");
        portIdentityFile = new PortParameter(this, "identity");
        portCommand = new TypedIOPort(this, "command", true, false);
        portStdout= new TypedIOPort(this, "stdout", false, true);
        portStderr = new TypedIOPort(this, "stderr", false, true);
        portReturnValue = new TypedIOPort(this, "return", false, true);
        portErrors = new TypedIOPort(this, "errors", false, true);

        // Set the type constraints.
        portUser.setTypeEquals(BaseType.STRING);
        portHost.setTypeEquals(BaseType.STRING);
        portIdentityFile.setTypeEquals(BaseType.STRING);
        portCommand.setTypeEquals(BaseType.STRING);
        portStdout.setTypeEquals(BaseType.STRING);
        portStderr.setTypeEquals(BaseType.STRING);
        portReturnValue.setTypeEquals(BaseType.INT);
        portErrors.setTypeEquals(BaseType.STRING);

        _attachText(
            "_iconDescription",
            "<svg>\n"
            + "<rect x=\"0\" y=\"0\" "
            + "width=\"75\" height=\"50\" style=\"fill:gray\"/>\n"
            + "<text x=\"5\" y=\"30\""
            + "style=\"font-size:25; fill:yellow; font-family:SansSerif\">"
            + "SSH2</text>\n"
            + "</svg>\n");
    }

    ////////////////// Public ports and parameters ///////////////////////

    public PortParameter portUser;
    public PortParameter portHost;
    public PortParameter portIdentityFile;
    public TypedIOPort portCommand;

    public TypedIOPort portStdout;
    public TypedIOPort portStderr;
    public TypedIOPort portReturnValue;
    public TypedIOPort portErrors;


    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////


    /** Callback for changes in attribute values
     *  Get the WSDL from the given URL.
     *
     *  @param a The attribute that changed.
     *  @exception IllegalActionException
     */
    public void attributeChanged (Attribute at) throws IllegalActionException {
      if ((at == portUser) || (at == portHost)) {
        if (! (portUser.getExpression().equals(""))) {
          String temp1 = portUser.getExpression();
          if (! (portHost.getExpression().equals(""))) {
            String temp2 = ( (StringToken) portHost.getToken()).stringValue();
            _attachText(
                "_iconDescription",
                "<svg>\n"
                + "<rect x=\"0\" y=\"0\" "
                + "width=\"90\" height=\"60\" style=\"fill:gray\"/>\n"
                + "<text x=\"5\" y=\"15\""
                + "style=\"font-size:12; fill:yellow; font-family:SansSerif\">"
                + temp1 + "\n@\n"
                + temp2 + "</text>\n"
                + "</svg>\n");
          }
          else {
            _attachText(
                "_iconDescription",
                "<svg>\n"
                + "<rect x=\"0\" y=\"0\" "
                + "width=\"90\" height=\"60\" style=\"fill:gray\"/>\n"
                + "<text x=\"5\" y=\"15\""
                + "style=\"font-size:12; fill:yellow; font-family:SansSerif\">"
                + temp1 + "\n@\nunknown_host</text>\n"
                + "</svg>\n");
          }
        }
        else {
          if (portHost.getExpression().equals("")){
            _attachText(
                "_iconDescription",
                "<svg>\n"
                + "<rect x=\"0\" y=\"0\" "
                + "width=\"75\" height=\"50\" style=\"fill:gray\"/>\n"
                + "<text x=\"5\" y=\"30\""
                + "style=\"font-size:25; fill:yellow; font-family:SansSerif\">"
                + "SSH2</text>\n"
                + "</svg>\n");
          }
          else {
            String temp2 = ( (StringToken) portHost.getToken()).stringValue();
            _attachText(
                "_iconDescription",
                "<svg>\n"
                + "<rect x=\"0\" y=\"0\" "
                + "width=\"90\" height=\"60\" style=\"fill:gray\"/>\n"
                + "<text x=\"5\" y=\"15\""
                + "style=\"font-size:12; fill:yellow; font-family:SansSerif\">"
                + "unknown_user\n@\n"
                + temp2 + "</text>\n"
                + "</svg>\n");
          }
        }

      }
    } //end-of-attributeChanged

    /** Send the token in the <i>value</i> parameter to the output.
     *  @exception IllegalActionException If it is thrown by the
     *   send() method sending out the token.
     */
    public void fire() throws IllegalActionException {
        super.fire();

        portUser.update();
        portHost.update();
        portIdentityFile.update();

        String strUser = ( (StringToken) portUser.getToken()).stringValue();
        String strHost = ( (StringToken) portHost.getToken()).stringValue();
        String strIdentity = ( (StringToken) portIdentityFile.getToken()).stringValue();
        String strCommand = ( (StringToken) portCommand.get(0)).stringValue();

        try {
            _connect(strUser, strHost, strIdentity);
            _exec(strCommand);
        } catch(IllegalActionException e) {
            // caught an exception, so output it to the errors port
            portStdout.send(0, new StringToken(""));
            portStderr.send(0, new StringToken(""));
            portReturnValue.send(0, new IntToken(0));
            portErrors.send(0, new StringToken(e.getMessage()));
        }
    }


    /** Terminate any sessions.
     *  This method is invoked exactly once per execution
     *  of an application.    None of the other action methods should be
     *  be invoked after it.
     *  @exception IllegalActionException Not thrown in this base class.
     */
    public void wrapup() throws IllegalActionException {
        _disconnect();
    }



    ////////////////// Private Methods ///////////////////////

    /**
     * @throws IllegalActionException If the connection fails.
     */
    private void _connect(String strUser, String strHost, String strIdentity)
        throws IllegalActionException {

        _debug("Connecting with " + strUser + "@" + strHost + " with identity: " + strIdentity);

        try {

            if(_setIdentities.add(strIdentity)) {
                // we haven't seen this identity before
                _jsch.addIdentity(strIdentity);
            }

            if(!strUser.equals(_strOldUser) || !strHost.equals(_strOldHost)) {
                if(null != _session) {
                    _disconnect();
                }

                _session = _jsch.getSession(strUser, strHost, 22);
                _strOldUser = strUser;
                _strOldHost = strHost;

                // username and passphrase will be given via UserInfo interface.
                UserInfo ui = new MyUserInfo();
                _session.setUserInfo(ui);
                _session.connect();
            }

        } catch(Exception e) {
            // a couple of possible exception messages that could happen here:
            // 1. java.io.FileNotFoundException
            // 2. session is down
            e.printStackTrace();
            throw new IllegalActionException("Exception caught in " +
                this.getFullName() + "\n(" + e.getClass().getName() + ")\n"
                + e.getMessage());
        }
    }

    /**
     *
     * @throws IllegalActionException if disconnect fails.
     */
    private void _disconnect() throws IllegalActionException {
        if(null == _session) {
            // no session, so nothing to disconnect
            return;
        }

        try {
            _session.disconnect();
        } catch(Exception e) {
            e.printStackTrace();
            throw new IllegalActionException("Exception caught in " +
                this.getFullName() + "\n(" + e.getClass().getName() + ")\n"
                + e.getMessage());
        }
    }

    /**
     *
     * @throws IllegalActionException
     */
    private void _exec(String execCommand) throws IllegalActionException {
        if(null == _session) {
            // no session, so way to execute
            return;
        }

        try {
            final ChannelExec channel = (ChannelExec) _session.openChannel("exec");
            channel.setCommand(execCommand);

            ByteArrayOutputStream streamOut = new ByteArrayOutputStream();
            ByteArrayOutputStream streamErr = new ByteArrayOutputStream();

            channel.setOutputStream(streamOut);
            channel.setErrStream(streamErr);

            channel.connect();

            // wait for it to finish
            _thread =
                new Thread() {
                    public void run() {
                        while (!channel.isEOF()) {
                            if (_thread == null) {
                                return;
                            }
                            try {
                                sleep(500);
                            }
                            catch (Exception e) {
                                // ignored
                            }
                        }
                    }
                };

            _thread.start();
            _thread.join(0); // set this to a different value to time out

            if (_thread.isAlive()) {
                // ran out of time
                _thread = null;
                throw new IllegalActionException("In " +
                    this.getFullName() + ": Remote operation timed out!");
            }
            // completed successfully

            // this is the wrong test if the remote OS is OpenVMS,
            // but there doesn't seem to be a way to detect it.
            int ec = channel.getExitStatus();
            channel.disconnect();

            portStdout.send(0, new StringToken(streamOut.toString()));
            portStderr.send(0, new StringToken(streamErr.toString()));
            portReturnValue.send(0, new IntToken(ec));
            portErrors.send(0, new StringToken(""));

        } catch (Exception e) {
            e.printStackTrace();
            throw new IllegalActionException("Exception caught in " +
                this.getFullName() + "\n(" + e.getClass().getName() + ")\n"
                + e.getMessage());
        }
    }



    ////////////////// Private variables ///////////////////////

    private JSch _jsch = null;
    private Session _session = null;
    private Thread _thread = null;
    private HashSet _setIdentities = null;
    private String _strOldUser = null;
    private String _strOldHost = null;

    ////////////////// Inner classes ///////////////////////

    public static class MyUserInfo
            implements UserInfo {

        public String getPassword() {
            return null;
        }

        public boolean promptYesNo(String str) {
            // This method gets called to answer the question similar to
            // "are you sure you want to connect to host whose key
            // is not in database ..."
            return true;
        }

        public String getPassphrase() {
            return null;
        }

        public boolean promptPassphrase(String message) {
            return false;
        }

        public boolean promptPassword(String message) {
            return true;
        }

        public void showMessage(String message) {
            // This method gets called when the server sends over a MOTD.
            //JOptionPane.showMessageDialog(null, message);
        }
    }

}

// vim: sw=4 ts=4 et
